#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool calendaring objects.
"""
import base64

from persistent.dict import PersistentDict
from persistent import Persistent
from zope.interface import implements, implementer
from zope.schema import getFieldNames
from zope.component import adapts, adapter
from zope.annotation.interfaces import IAttributeAnnotatable, IAnnotations
from zope.container.contained import Contained
from zope.location.interfaces import ILocation

from schooltool.calendar.icalendar import read_icalendar
from schooltool.calendar.interfaces import ICalendar
from schooltool.calendar.interfaces import ICalendarEvent
from schooltool.calendar.interfaces import IExpandedCalendarEvent
from schooltool.calendar.mixins import CalendarMixin
from schooltool.calendar.simple import SimpleCalendarEvent
from schooltool.app.interfaces import ISchoolToolCalendarEvent
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import IWriteCalendar
from schooltool.app.interfaces import IHaveCalendar


CALENDAR_KEY = 'schooltool.app.calendar.Calendar'


class CalendarEvent(SimpleCalendarEvent, Persistent, Contained):
    """A persistent calendar event contained in a persistent calendar."""

    implements(ISchoolToolCalendarEvent, IAttributeAnnotatable)

    __parent__ = None

    resources = property(lambda self: self._resources)

    def __init__(self, *args, **kwargs):
        resources = kwargs.pop('resources', ())
        SimpleCalendarEvent.__init__(self, *args, **kwargs)
        self.__name__ = base64.encodestring(self.unique_id.encode('utf-8')).replace('\n', '')
        self._resources = ()
        for resource in resources:
            self.bookResource(resource)

    def __conform__(self, interface):
        if interface is ICalendar:
            return self.__parent__

    def bookResource(self, resource):
        calendar = ISchoolToolCalendar(resource)
        if resource in self.resources:
            raise ValueError('resource already booked')
        if calendar is self.__parent__:
            raise ValueError('cannot book itself')
        self._resources += (resource, )
        if self.__parent__ is not None:
            calendar.addEvent(self)

    def unbookResource(self, resource):
        if resource not in self.resources:
            raise ValueError('resource not booked')
        self._resources = tuple([r for r in self.resources
                                 if r is not resource])
        ISchoolToolCalendar(resource).removeEvent(self)

    @property
    def owner(self):
        if self.__parent__:
            return self.__parent__.__parent__
        else:
            return None


class Calendar(Persistent, CalendarMixin):
    """A persistent calendar."""

    # CalendarMixin is only used for the expand() method

    implements(ISchoolToolCalendar, IAttributeAnnotatable)

    __name__ = 'calendar'

    title = property(lambda self: self.__parent__.title)

    def __init__(self, owner):
        self.events = PersistentDict()
        self.__parent__ = owner

    def __iter__(self):
        return self.events.itervalues()

    def __len__(self):
        return len(self.events)

    def addEvent(self, event):
        assert ISchoolToolCalendarEvent.providedBy(event)
        if event.unique_id in self.events:
            raise ValueError('an event with this unique_id already exists')
        if event.__parent__ is None:
            for resource in event.resources:
                if ISchoolToolCalendar(resource) is self:
                    raise ValueError('cannot book itself')
            event.__parent__ = self
            for resource in event.resources:
                ISchoolToolCalendar(resource).addEvent(event)
        elif self.__parent__ not in event.resources:
            raise ValueError("Event already belongs to a calendar")
        self.events[event.unique_id] = event

    def removeEvent(self, event):
        if self.__parent__ in event.resources:
            event.unbookResource(self.__parent__)
        else:
            del self.events[event.unique_id]
            parent_calendar = event.__parent__
            if self is parent_calendar:
                for resource in event.resources:
                    event.unbookResource(resource)
                event.__parent__ = None

    def clear(self):
        # clear is not actually used anywhere in schooltool.app (except tests),
        # so it doesn't have to be efficient.
        for e in list(self):
            self.removeEvent(e)

    def find(self, unique_id):
        return self.events[unique_id]


def getCalendar(owner):
    """Adapt an ``IAnnotatable`` object to ``ISchoolToolCalendar``."""
    annotations = IAnnotations(owner)
    try:
        return annotations[CALENDAR_KEY]
    except KeyError:
        calendar = Calendar(owner)
        annotations[CALENDAR_KEY] = calendar
        return calendar
getCalendar.factory = Calendar # Convention to make adapter introspectable


class WriteCalendar(object):
    r"""An adapter that allows writing iCalendar data to a calendar.

        >>> calendar = Calendar(None)
        >>> adapter = WriteCalendar(calendar)
        >>> adapter.write('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... SUMMARY:LAN party
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... ''')
        >>> for e in calendar:
        ...     print e.dtstart.strftime('%Y-%m-%d %H:%M'), e.title
        2005-02-26 16:00 LAN party

    Supporting other charsets would be nice too:

        >>> calendar = Calendar(None)
        >>> adapter = WriteCalendar(calendar)
        >>> adapter.write('''\
        ... BEGIN:VCALENDAR
        ... VERSION:2.0
        ... PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        ... BEGIN:VEVENT
        ... UID:some-random-uid@example.com
        ... SUMMARY:LAN party %s
        ... DTSTART:20050226T160000
        ... DURATION:PT6H
        ... DTSTAMP:20050203T150000
        ... END:VEVENT
        ... END:VCALENDAR
        ... ''' %  chr(163), charset='latin-1')
        >>> titles = [e.title for e in calendar]
        >>> titles[0]
        u'LAN party \xa3'

    """

    adapts(ISchoolToolCalendar)
    implements(IWriteCalendar)

    # Hook for unit tests.
    read_icalendar = staticmethod(read_icalendar)

    _event_attrs = getFieldNames(ICalendarEvent)

    def __init__(self, context, request=None):
        self.calendar = context

    def write(self, data, charset='UTF-8'):
        changes = {} # unique_id -> (old_event, new_event)
        for e in self.calendar:
            changes[e.unique_id] = (e, None)

        for event in self.read_icalendar(data, charset):
            old_event = changes.get(event.unique_id, (None, ))[0]
            changes[event.unique_id] = (old_event, event)

        for old_event, new_event in changes.itervalues():
            if old_event is None:
                # new_event is a SimpleCalendarEvent, we need a CalendarEvent
                kwargs = dict([(attr, getattr(new_event, attr))
                               for attr in self._event_attrs])
                self.calendar.addEvent(CalendarEvent(**kwargs))
            elif new_event is None:
                self.calendar.removeEvent(old_event)
            elif old_event != new_event:
                # modify in place
                for attr in self._event_attrs:
                    setattr(old_event, attr, getattr(new_event, attr))


def clearCalendarOnDeletion(event):
    """When you delete an object, it's calendar should be cleared

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.testing.setup import setUpCalendaring

        >>> setUp()
        >>> setUpCalendaring()

        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> from schooltool.app.cal import clearCalendarOnDeletion
        >>> zope.event.subscribers.append(clearCalendarOnDeletion)

    We will need some object that implements IHaveCalendar for that:

        >>> from zope.container.btree import BTreeContainer
        >>> container = BTreeContainer()
        >>> from schooltool.person.person import Person
        >>> container = BTreeContainer()
        >>> container['petras'] = petras =  Person(username="Petras")
        >>> def clearCalendar():
        ...     print "Clearing calendar"
        >>> ISchoolToolCalendar(petras).clear = clearCalendar

    If we delete Petras his calendar should be cleared:

        >>> del container['petras']
        Clearing calendar

    Restore old subscribers:

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """
    if IHaveCalendar.providedBy(event.object):
        ISchoolToolCalendar(event.object).clear()


@adapter(IExpandedCalendarEvent)
@implementer(ILocation)
def expandedEventLocation(event):
    original = event.__dict__['original']
    return ILocation(original)



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact objects
"""
from persistent import Persistent

from zope.catalog.text import TextIndex
from zope.component import adapter, adapts
from zope.index.text.interfaces import ISearchableText
from zope.interface import implementer
from zope.interface import implements
from zope.intid.interfaces import IIntIds
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer
from zope.component import getUtility

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.catalog import AttributeCatalog
from schooltool.utility.utility import UtilitySetUp
from schooltool.contact.interfaces import IContactPersonInfo
from schooltool.contact.interfaces import IContactable
from schooltool.contact.interfaces import IContact, IContactContained
from schooltool.contact.interfaces import IContactContainer
from schooltool.contact.interfaces import IUniqueFormKey
from schooltool.relationship.uri import URIObject
from schooltool.relationship.relationship import BoundRelationshipProperty
from schooltool.relationship.relationship import RelationshipProperty
from schooltool.relationship.relationship import RelationshipSchema
from schooltool.securitypolicy import crowds
from schooltool.table.catalog import ConvertingIndex
from schooltool.common import simple_form_key
from schooltool.person.interfaces import IPerson
from schooltool.course.section import PersonInstructorsCrowd

from schooltool.common import SchoolToolMessage as _


URIContactRelationship = URIObject('http://schooltool.org/ns/contact',
                                   'Contact relationship',
                                   'The contact relationship.')

URIContact = URIObject('http://schooltool.org/ns/contact/contact',
                       'Contact', 'A contact relationship contact record role.')

URIPerson = URIObject('http://schooltool.org/ns/contact/person',
                      'Person', 'A contact relationship person role.')

Contact = RelationshipSchema(URIContactRelationship,
                             contact=URIContact, person=URIPerson)


class ContactGroup(crowds.DescriptionGroup):

    title = _(u"Contacts")

    description = _(u''
        'All SchoolTool users have contact information, that includes home '
        'address, phone number, email address, etc.'
        '\n'
        '"External contacts" can be created - people who are not SchoolTool '
        'users, but may need to recieve notifications and information about '
        'the student or the school as a whole.')


class ContactContainer(BTreeContainer):
    """Container of persons."""

    implements(IContactContainer)


class Contact(Persistent, Contained):
    """Contact."""

    implements(IContactContained)

    prefix = None
    first_name = None
    middle_name = None
    last_name = None
    suffix = None
    photo = None
    address_line_1 = None
    address_line_2 = None
    city = None
    state = None
    country = None
    postal_code = None
    email = None
    home_phone = None
    work_phone = None
    mobile_phone = None
    other_1 = None
    other_2 = None
    language = None

    persons = RelationshipProperty(URIContactRelationship,
                                   my_role=URIContact,
                                   other_role=URIPerson)

    @property
    def title(self):
        return "%s %s" % (self.first_name, self.last_name)


class ContactPersonInfo(Persistent, Contained):
    """Additional information about contact of a specific person."""

    implements(IContactPersonInfo)

    __parent__ = None
    relationship = None

    def getRelationshipTitle(self):
        vocabulary = IContactPersonInfo['relationship'].vocabulary
        try:
            term = vocabulary.getTerm(self.relationship)
        except LookupError:
            return u''
        return term.title


class ContextRelationshipProperty(RelationshipProperty):
    """Context relationship property."""

    def __get__(self, instance, owner):
        """Bind the property to the context of an instance."""
        if instance is None:
            return self
        else:
            return BoundRelationshipProperty(instance.context,
                                             self.rel_type,
                                             self.my_role,
                                             self.other_role)


class Contactable(object):
    implements(IContactable)

    contacts = ContextRelationshipProperty(URIContactRelationship,
                                           my_role=URIPerson,
                                           other_role=URIContact)

    def __init__(self, context):
        self.context = context


class ContactAppStartup(StartUpBase):
    def __call__(self):
        if 'schooltool.contact' not in self.app:
            self.app['schooltool.contact'] = ContactContainer()


class ContactInit(InitBase):
    def __call__(self):
        self.app['schooltool.contact'] = ContactContainer()


@implementer(IContactContainer)
@adapter(ISchoolToolApplication)
def getContactContainer(app):
    return app['schooltool.contact']


@adapter(IContact)
@implementer(IUniqueFormKey)
def getContactFormKey(contact):
    doc_id = getUtility(IIntIds).getId(contact)
    return 'contacts.%s%x' % (simple_form_key(contact), doc_id)


class ContactCatalog(AttributeCatalog):
    version = '4 - added text index'
    interface = IContact
    attributes = ('first_name', 'last_name', 'title')

    def setIndexes(self, catalog):
        super(ContactCatalog, self).setIndexes(catalog)
        catalog['form_keys'] = ConvertingIndex(converter=IUniqueFormKey)
        catalog['text'] = TextIndex('getSearchableText', ISearchableText, True)


getContactCatalog = ContactCatalog.get


class SearchableTextContact(object):

    adapts(IContact)
    implements(ISearchableText)

    def __init__(self, context):
        self.context = context

    def getSearchableText(self):
        result = [self.context.first_name, self.context.last_name]
        return ' '.join(result)


class ContactPersonInstructorsCrowd(PersonInstructorsCrowd):
    """Crowd of instructors of any of related persons to this contact."""

    title = _(u'Instructors')
    description = _(u'Instructors of student(s) related to this contact.')

    def contains(self, principal):
        user = IPerson(principal, None)
        for person in self.context.persons:
            sections = self._getSections(person)
            for section in sections:
                if user in section.instructors:
                    return True
        return False



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.timetable.timetable
"""
import doctest
import pytz
import unittest
from pprint import pprint
from datetime import date, time, datetime, timedelta

import zope.event
from zope.app.testing import setup
from zope.intid.interfaces import IIntIds
from zope.component import provideUtility
from zope.container.contained import containedEvent

from schooltool.common import DateRange
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import DayTemplateSchedule
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Meeting
from schooltool.timetable.schedule import Period
from schooltool.timetable.tests import ScheduleStub
from schooltool.timetable.timetable import Timetable

from schooltool.timetable.schedule import iterMeetingsInTimezone


class OneTemplateSchedule(DayTemplateSchedule):

    def initTemplates(self):
        super(OneTemplateSchedule, self).initTemplates()
        self.templates['default'] = DayTemplate('Day 1')

    @property
    def default(self):
        return self.templates['default']

    def iterDates(self, dates):
        schedule = self.__parent__
        scheduled_dates = DateRange(schedule.first, schedule.last)
        for cursor in dates:
            if cursor not in scheduled_dates:
                yield None
            else:
                yield self.default


class TimetableForTests(Timetable):

    def setUp(self, periods=(), time_slots=()):
        self.periods, event = containedEvent(
            OneTemplateSchedule(), self, 'periods')
        zope.event.notify(event)
        self.periods.initTemplates()
        self.time_slots, event = containedEvent(
            OneTemplateSchedule(), self, 'time_slots')
        zope.event.notify(event)
        self.time_slots.initTemplates()
        for n, title in enumerate(periods):
            self.periods.default['%d' % (n+1)] = Period(title=title)
        for n, tstart in enumerate(time_slots):
            duration = timedelta(0, 900)
            self.time_slots.default['%d' % (n+1)] = TimeSlot(tstart, duration)


def print_utc_times(meetings):
    for meeting in meetings:
        start_time_s = meeting.dtstart.strftime('%Y-%m-%d %H:%M %Z')
        utc_time = meeting.dtstart.astimezone(pytz.UTC)
        utc_time_s = utc_time.strftime('%Y-%m-%d %H:%M %Z')
        print start_time_s, '->', utc_time_s


def test_Timetable_timezones():
    """Tests for Timetable.

    Let's start with a timetable with a meeting every hour every day.

        >>> tt = TimetableForTests(
        ...     date(2011, 10, 28), date(2011, 11, 01), timezone='UTC')
        >>> periods = [chr(ord('A')+x) for x in range(24)]
        >>> time_slots = [time(x, 00) for x in range(24)]
        >>> tt.setUp(
        ...    periods=periods,
        ...    time_slots=time_slots)

        >>> pprint(list(tt.iterMeetings(tt.first, tt.last)))
        [<Meeting A on 2011-10-28 00:00 UTC>,
         <Meeting B on 2011-10-28 01:00 UTC>,
         <Meeting C on 2011-10-28 02:00 UTC>,
         ...
         <Meeting W on 2011-10-28 22:00 UTC>,
         <Meeting X on 2011-10-28 23:00 UTC>,
         <Meeting A on 2011-10-29 00:00 UTC>,
         ...
         <Meeting X on 2011-11-01 23:00 UTC>]

        >>> len(list(tt.iterMeetings(tt.first)))
        24

        >>> len(list(tt.iterMeetings(tt.first, tt.last)))
        120

     If the timetable is in a timezone and dst shift occurs, we can
     observe it:

        >>> tt.timezone = 'Europe/Vilnius'
        >>> pprint(list(tt.iterMeetings(tt.first, tt.last)))
        [<Meeting A on 2011-10-28 00:00 EEST>,
         ...
         <Meeting A on 2011-10-30 00:00 EEST>,
         <Meeting B on 2011-10-30 01:00 EEST>,
         <Meeting C on 2011-10-30 02:00 EEST>,
         <Meeting D on 2011-10-30 03:00 EET>,
         <Meeting E on 2011-10-30 04:00 EET>,
         <Meeting F on 2011-10-30 05:00 EET>,
         ...
         <Meeting X on 2011-11-01 23:00 EET>]

        >>> len(list(tt.iterMeetings(tt.first, tt.last)))
        120

        >>> print_utc_times(tt.iterMeetings(tt.first, tt.last))
        2011-10-28 00:00 EEST -> 2011-10-27 21:00 UTC
        2011-10-28 01:00 EEST -> 2011-10-27 22:00 UTC
        ...
        2011-10-30 02:00 EEST -> 2011-10-29 23:00 UTC
        2011-10-30 03:00 EET -> 2011-10-30 01:00 UTC
        ...
        2011-11-01 23:00 EET -> 2011-11-01 21:00 UTC

    We can look at the timetable from another timezone.  Note that
    we get 3 less meetings for the two dates, because first date
    2011-10-28 in UTC starts 3 hours later than in Vilnius.

        >>> pprint(list(iterMeetingsInTimezone(
        ...     tt, 'UTC', tt.first, date(2011, 10, 30))))
        [<Meeting D on 2011-10-28 03:00 EEST>,
         <Meeting E on 2011-10-28 04:00 EEST>,
         ...
         <Meeting B on 2011-10-30 01:00 EEST>,
         <Meeting C on 2011-10-30 02:00 EEST>,
         <Meeting D on 2011-10-30 03:00 EET>,
         <Meeting E on 2011-10-30 04:00 EET>,
         ...
         <Meeting X on 2011-10-30 23:00 EET>,
         <Meeting A on 2011-10-31 00:00 EET>,
         <Meeting B on 2011-10-31 01:00 EET>]

    Timetable exception dates are defined in timetable's timezone:

        >>> exception_starts = pytz.timezone(tt.timezone).localize(
        ...     datetime(2011, 10, 29, 11, 15))
        >>> ex_meeting = Meeting(exception_starts, timedelta(0, 900))
        >>> tt.exceptions[date(2011, 10, 29)] = [ex_meeting]

        >>> pprint(list(iterMeetingsInTimezone(
        ...     tt, 'UTC', tt.first, tt.last)))
        [<Meeting D on 2011-10-28 03:00 EEST>,
         <Meeting E on 2011-10-28 04:00 EEST>,
         ...
         <Meeting W on 2011-10-28 22:00 EEST>,
         <Meeting X on 2011-10-28 23:00 EEST>,
         <Meeting on 2011-10-29 11:15 EEST>,
         <Meeting A on 2011-10-30 00:00 EEST>,
         <Meeting B on 2011-10-30 01:00 EEST>,
         ...
         <Meeting X on 2011-11-01 23:00 EET>]

    """


def test_Timetable_duplicate_time_slot():
    """Tests for Timetable that falls into duplicate hour when
    clocks are adjusted forward in a timezone.

    Let's start with a timetable with a meeting every hour every day.

        >>> tt = TimetableForTests(
        ...     date(2011, 3, 26), date(2011, 3, 28),
        ...     timezone='Europe/Vilnius')
        >>> periods = [chr(ord('A')+x) for x in range(24)]
        >>> time_slots = [time(x, 00) for x in range(24)]
        >>> tt.setUp(
        ...    periods=periods,
        ...    time_slots=time_slots)

    Time period 03:00-04:00 does not exist in given timezone in 2011-03-27
    as a result of clocks adjusted forward.  SchoolTool timetables schedule
    the 03:xx and 04:xx entries at the same (adjusted) hour, so as not to
    lose planned events.

        >>> print_utc_times(tt.iterMeetings(tt.first, tt.last))
        2011-03-26 00:00 EET -> 2011-03-25 22:00 UTC
        ...
        2011-03-27 02:00 EET -> 2011-03-27 00:00 UTC
        2011-03-27 03:00 EET -> 2011-03-27 01:00 UTC
        2011-03-27 04:00 EEST -> 2011-03-27 01:00 UTC
        2011-03-27 05:00 EEST -> 2011-03-27 02:00 UTC
        ...

    """


def setUp(test=None):
    setup.placelessSetUp()
    provideUtility(object(), IIntIds)


def tearDown(test=None):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown),
           ])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Implementation of relationships for IAnnotatable objects.

Relationships are represented as collections of links.  A link defines one
half of a relationship.  The storage of links on an object is determined by
an IRelationshipLinks adapter.  There is a default adapter registered for
all IAnnotatable objects that uses Zope 3 annotations.
"""

from zope.annotation.interfaces import IAnnotations
from schooltool.relationship.relationship import LinkSet


def getRelationshipLinks(context):
    """Adapt an IAnnotatable object to IRelationshipLinks."""
    annotations = IAnnotations(context)
    key = 'schooltool.relationship.RelationshipLinks'
    try:
        return annotations[key]
    except KeyError:
        annotations[key] = LinkSet()
        annotations[key].__parent__ = context
        annotations[key].__name__ = "relationships"
        return annotations[key]



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve34
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.container.btree import BTreeContainer

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.email.interfaces import IEmailContainer


class AppStub(BTreeContainer):
    implements(ISchoolToolApplication)


class EmailContainerStub(BTreeContainer):
    implements(IEmailContainer)
    enabled = None
    hostname = None


def doctest_evolve34():
    """Test evolution to generation 34.

    Set some apps:

        >>> from schooltool.generations.tests import ContextStub
        >>> context = ContextStub()
        >>> context.root_folder['app1'] = app1 = AppStub()
        >>> context.root_folder['app2'] = app2 = AppStub()
        >>> context.root_folder['app3'] = app3 = AppStub()

    And email containers in them:

        >>> from schooltool.email.mail import EMAIL_KEY
        >>> container1 = app1[EMAIL_KEY] = EmailContainerStub()
        >>> container2 = app2[EMAIL_KEY] = EmailContainerStub()
        >>> container3 = app3[EMAIL_KEY] = EmailContainerStub()

    Set some 'hostname' attributes:

        >>> container1.hostname = ''
        >>> container3.hostname = 'localhost'

    Check the 'enabled' attributes:

        >>> getattr(container1, 'enabled', None) is None
        True
        >>> getattr(container2, 'enabled', None) is None
        True
        >>> getattr(container3, 'enabled', None) is None
        True

    Evolve:

        >>> from zope.component import provideAdapter
        >>> from schooltool.email.mail import getEmailContainer
        >>> provideAdapter(getEmailContainer, [ISchoolToolApplication],
        ...                IEmailContainer)
        >>> from schooltool.generations.evolve34 import evolve
        >>> evolve(context)

    Check the 'hostname' and 'enabled' attributes:

        >>> container1.hostname
        ''
        >>> container1.enabled
        False

        >>> getattr(container2, 'hostname', None) is None
        True
        >>> container2.enabled
        False

        >>> container3.hostname
        'localhost'
        >>> container3.enabled
        True

    """


def setUp(test):
    setup.placelessSetUp()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Helpers for the evolution.
"""
from ZODB.broken import Broken


def assert_not_broken(*objects):
    for obj in objects:
        broken_list = []
        if isinstance(obj, Broken):
            broken_list.append(obj)
        try:
            attrs = sorted(set(list(obj.__dict__) +
                               list(obj.__class__.__dict__)))
        except:
            attrs = []
        finally:
            for name in attrs:
                a = getattr(obj, name)
                if isinstance(a, Broken):
                    broken_list.append((obj, name, a))
        assert not broken_list, broken_list


class BuildContext(object):
    _options = None
    def __init__(self, *args, **kw):
        self._options = {}
        self.update(*args, **kw)

    def __getattr__(self, name):
        if name in self._options:
            return self._options[name]
        raise AttributeError(name)

    def update(self, *args, **options):
        if len(args) == 1 and isinstance(args[0], BuildContext):
            self._options.update(args[0]._options)
        self._options.update(options)

    def __call__(self, *args, **options):
        new_context = self.__class__(**self._options)
        new_context.update(*args, **options)
        return new_context




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool timetabling views.
"""
import datetime
import urllib
import pytz
from persistent.list import PersistentList

import zope.schema
import zope.event
import zope.lifecycleevent
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import getUtility, queryMultiAdapter
from zope.intid.interfaces import IIntIds
from zope.publisher.browser import BrowserView
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from z3c.form import form, button

import schooltool.skin.flourish.breadcrumbs
import schooltool.skin.flourish.interfaces
import schooltool.skin.flourish.form
import schooltool.skin.flourish.page
from schooltool.calendar.utils import parse_date, parse_time
from schooltool.common.inlinept import InheritTemplate
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool.term.interfaces import ITerm
from schooltool.term.term import getTermForDate
from schooltool.timetable.schedule import MeetingException
from schooltool.timetable.interfaces import IHaveSchedule

from schooltool.common import SchoolToolMessage as _


class SpecialDayView(BrowserView):
    """The view for changing the periods for a particular day.

    The typical use case: some periods get shortened or and some get
    cancelled altogether if some special event is held at noon.
    """

    select_template = ViewPageTemplateFile('templates/specialday_select.pt')
    form_template = ViewPageTemplateFile('templates/specialday_change.pt')

    error = None
    field_errors = None
    field_error_message = _('Some values were invalid.'
                            '  They are highlighted in red.')
    date = None
    term = None

    def delta(self, start, end):
        """
        Returns a timedelta between two times

            >>> from datetime import time, timedelta
            >>> view = SpecialDayView(None, None)
            >>> view.delta(time(11, 10), time(12, 20))
            datetime.timedelta(0, 4200)

        If a result is negative, it is 'wrapped around':

            >>> view.delta(time(11, 10), time(10, 10)) == timedelta(hours=23)
            True
        """
        today = datetime.date.today()
        dtstart = datetime.datetime.combine(today, start)
        dtend = datetime.datetime.combine(today, end)
        delta = dtend - dtstart
        if delta < datetime.timedelta(0):
            delta += datetime.timedelta(1)
        return delta

    @property
    def schedule(self):
        return self.context

    def meeting_form_key(self, meeting):
        mid = meeting.meeting_id or ''
        mid = unicode(meeting.meeting_id).encode('punycode')
        mid = urllib.quote(mid)
        time = '%s:%s' % (meeting.dtstart.hour, meeting.dtstart.minute)
        time_key = '%s.%s.%s' % (time, meeting.duration, mid)
        period = meeting.period
        if period is None:
            return time_key
        int_ids = getUtility(IIntIds)
        return '%s.%s' % (int_ids.getId(period), mid)

    def getMeetings(self):
        meeting_info = {}
        for meeting in self.schedule.iterMeetings(self.date):
            meeting = removeSecurityProxy(meeting)
            start_time = meeting.dtstart.time()
            end_time = self.timeplustd(start_time, meeting.duration)
            form_key = self.meeting_form_key(meeting)
            meeting_info[form_key] = {
                'form_key': form_key,
                'meeting': meeting,
                'orig_start_time': '',
                'orig_end_time': '',
                'start_time': self.request.get('%s_start' % form_key,
                                               start_time.strftime("%H:%M")),
                'end_time': self.request.get('%s_end' % form_key,
                                             end_time.strftime("%H:%M")),
                }

        for meeting in self.schedule.iterOriginalMeetings(self.date):
            meeting = removeSecurityProxy(meeting)
            form_key = self.meeting_form_key(meeting)
            start_time = meeting.dtstart.time()
            end_time = self.timeplustd(start_time, meeting.duration)
            if form_key in meeting_info:
                info = meeting_info[form_key]
                info['orig_start_time'] = start_time.strftime("%H:%M")
                info['orig_end_time'] = end_time.strftime("%H:%M")
            else:
                meeting_info[form_key] = {
                    'form_key': form_key,
                    'meeting': meeting,
                    'orig_start_time': start_time.strftime("%H:%M"),
                    'orig_end_time': end_time.strftime("%H:%M"),
                    'start_time': self.request.get('%s_start' % form_key, ''),
                    'end_time': self.request.get('%s_end' % form_key, ''),
                }

        return sorted(meeting_info.values(),
                      key=lambda i: i['meeting'].dtstart)

    def extractMeeting(self, form_key):
        start_name = '%s_start' % form_key
        end_name = '%s_end' % form_key
        if (start_name in self.request and end_name in self.request
            and (self.request[start_name] or self.request[end_name])):
            start = end = None
            try:
                start = parse_time(self.request[start_name])
            except ValueError:
                pass
            try:
                end = parse_time(self.request[end_name])
            except ValueError:
                pass
            if start is None:
                self.field_errors.append(start_name)
            if end is None:
                self.field_errors.append(end_name)
            elif start is not None:
                duration = self.delta(start, end)
                return (start, duration)
        return None

    def extractMeetings(self):
        replacements = {}
        for meeting in self.schedule.iterOriginalMeetings(self.date):
            meeting = removeSecurityProxy(meeting)
            form_key = self.meeting_form_key(meeting)
            timespan = self.extractMeeting(form_key)
            if timespan is not None:
                replacements[form_key] = (meeting, ) + timespan

        for meeting in self.schedule.iterMeetings(self.date):
            meeting = removeSecurityProxy(meeting)
            form_key = self.meeting_form_key(meeting)
            timespan = self.extractMeeting(form_key)
            if timespan is not None:
                replacements[form_key] = (meeting, ) + timespan

        return replacements.values()

    def updateExceptions(self, replacements):
        template = PersistentList()
        tz = pytz.timezone(self.schedule.timezone)
        for meeting, start, duration in sorted(replacements,
                                               key=lambda r: r[1]):
            dtstart = datetime.datetime.combine(self.date, start)
            dtstart = dtstart.replace(tzinfo=tz)
            template.append(MeetingException(
                dtstart, duration,
                period=meeting.period,
                meeting_id=meeting.meeting_id))
        # XXX: broken permissions with PersistentDict
        exceptions = removeSecurityProxy(self.schedule.exceptions)
        exceptions[self.date] = template
        zope.lifecycleevent.modified(self.schedule)

    def update(self):
        """Read and validate form data, and update model if necessary.

        Also choose the correct template to render.
        """
        self.field_errors = []
        self.template = self.select_template

        if 'CANCEL' in self.request:
            self.request.response.redirect(
                absoluteURL(self.context, self.request))
            return

        if 'date' in self.request:
            try:
                self.date = parse_date(self.request['date'])
            except ValueError:
                self.error = _("Invalid date. Please use YYYY-MM-DD format.")
            else:
                self.term = getTermForDate(self.date)
                if self.term is None:
                    self.error = _("The date does not belong to any term.")
                    self.date = None

        if self.date:
            self.template = self.form_template

        if self.date and 'SUBMIT' in self.request:
            replacements = self.extractMeetings()
            if self.field_errors:
                self.error = self.field_error_message
            else:
                self.updateExceptions(replacements)
                self.request.response.redirect(
                    absoluteURL(self.context, self.request))

    def timeplustd(self, t, td):
        """Add a timedelta to time.

        datetime authors are cowards.

            >>> view = SpecialDayView(None, None)
            >>> from datetime import time, timedelta
            >>> view.timeplustd(time(10,0), timedelta(0, 5))
            datetime.time(10, 0, 5)
            >>> view.timeplustd(time(23,0), timedelta(0, 3660))
            datetime.time(0, 1)
        """
        dt = datetime.datetime.combine(datetime.date.today(), t)
        dt += td
        return dt.time()


    def __call__(self):
        self.update()
        return self.template()


class FlourishSpecialDayView(flourish.page.Page, SpecialDayView):
    select_template = ViewPageTemplateFile('templates/f_specialday_select.pt')
    form_template = ViewPageTemplateFile('templates/f_specialday_change.pt')
    flourish_template = InheritTemplate(flourish.page.Page.template)

    field_error_message = _('Please correct the marked fields below.')

    def update(self):
        SpecialDayView.update(self)
        self.content_template = self.template
        self.template = self.flourish_template


class ScheduleContainerView(BrowserView):
    template = ViewPageTemplateFile('templates/schedule_container_view.pt')

    @property
    def owner(self):
        return IHaveSchedule(self.context)

    @property
    def term(self):
        return ITerm(self.owner, None)

    @property
    def school_year(self):
        term = self.term
        if term is None:
            return None
        return ISchoolYear(term, None)

    def schedules(self):
        for name, schedule in list(self.context.items()):
            if schedule.timetable.__parent__ is not None:
                yield schedule
            else:
                # Remove stray schedule (LP: #1280528)
                del self.context[name]

    def timetables(self):
        snippets = filter(None, [
            queryMultiAdapter((schedule, self.request), name='schedule_table')
            for schedule in self.schedules()])
        return snippets

    def __call__(self):
        return self.template()


class FlourishScheduleContainerView(flourish.page.WideContainerPage,
                                    ScheduleContainerView):
    pass


class ScheduleDeleteView(BrowserView):
    template = ViewPageTemplateFile('templates/confirm_schedule_delete.pt')

    @property
    def schedule(self):
        name = self.request.get('schedule')
        if name is None or name not in self.context:
            return None
        return self.context[name]

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def __call__(self):
        schedule = self.schedule
        if schedule is not None:
            if 'CONFIRM' in self.request:
                del self.context[schedule.__name__]
                self.request.response.redirect(self.nextURL())
            elif 'CANCEL' in self.request:
                self.request.response.redirect(self.nextURL())
            else:
                return self.template()
        else:
            self.request.response.redirect(self.nextURL())


def scheduleOwnerTitle(context, request, view, name):
    owner = IHaveSchedule(context)
    return flourish.content.queryContentProvider(
        owner, request, view, 'title')


class ScheduleActionsLinks(flourish.page.RefineLinksViewlet):
    """Manager for Action links in scheduling views."""


class FlourishConfirmDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for confirming deletion of a timetable."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    def nextURL(self):
        link = flourish.content.queryContentProvider(
            self.context, self.request, self, 'done_link')
        if link is not None:
            return link.url
        return absoluteURL(self.context.__parent__, self.request)

    def delete(self):
        pass

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        next_url = self.nextURL()
        self.delete()
        self.request.response.redirect(next_url)
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishConfirmDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class ScheduleAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for schedule Add links."""


class ScheduleContainerBreadcrumbs(flourish.breadcrumbs.Breadcrumbs):

    @property
    def url(self):
        base_url = absoluteURL(self.crumb_parent, self.request)
        return '%s/schedule' % base_url



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool Testing Support
"""
__docformat__ = 'restructuredtext'
from schooltool.app.security import setUpLocalAuth
from schooltool.testing import registry

# ----------------------------- Session setup ------------------------------
from zope.publisher.interfaces import IRequest
from zope.component import provideAdapter, provideUtility
from zope.session.http import CookieClientIdManager
from zope.session.interfaces import ISessionDataContainer
from zope.session.interfaces import IClientIdManager, ISession
from zope.session.session import ClientId, Session
from zope.session.session import PersistentSessionDataContainer
def setUpSessions():
    """Set up the session machinery.

    Do this after placelessSetUp().
    """
    provideAdapter(ClientId)
    provideAdapter(Session, (IRequest,), ISession)
    provideUtility(CookieClientIdManager(), IClientIdManager)
    sdc = PersistentSessionDataContainer()
    provideUtility(sdc, ISessionDataContainer)


# --------------------- Create a SchoolTool application --------------------
from schooltool.app.app import SchoolToolApplication
def createSchoolToolApplication():
    """Create a ``SchoolToolApplication`` instance with all its high-level
    containers."""
    app = SchoolToolApplication()
    registry.setupApplicationContainers(app)
    return app


# ----------------- Setup SchoolTool application as a site -----------------
from zope.interface import Interface
from zope.interface import directlyProvides
from zope.component.hooks import setSite
from zope.site import LocalSiteManager
from zope.traversing.interfaces import IContainmentRoot
from schooltool.app.app import getSchoolToolApplication
from schooltool.app.security import PersonContainerAuthenticationPlugin
def setUpSchoolToolSite():
    """This should only be called after ``placefulSetUp()``."""
    app = createSchoolToolApplication()
    directlyProvides(app, IContainmentRoot)
    app.setSiteManager(LocalSiteManager(app))
    setUpLocalAuth(app)

    plugin = PersonContainerAuthenticationPlugin()
    provideUtility(plugin)
    provideAdapter(getSchoolToolApplication, (Interface,), ISchoolToolApplication)

    setSite(app)
    return app

# --------------- Setup Calendar Adapter and set IHaveCalendar -------------
from schooltool.app.interfaces import IHaveCalendar
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.cal import getCalendar
from schooltool.app.browser.cal import getCalendarEventDeleteLink
def setUpCalendaring():
    provideAdapter(getCalendar, (IHaveCalendar,), ISchoolToolCalendar)
    provideAdapter(getCalendarEventDeleteLink, name="delete_link")
    registry.setupCalendarComponents()


# ----------------- Setup SchoolTool application preferences ---------------
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import getApplicationPreferences
def setUpApplicationPreferences():
    """A utility method for setting up the ApplicationPreferences adapter."""
    provideAdapter(getApplicationPreferences,
                   (ISchoolToolApplication,), IApplicationPreferences)


# --------------------------------------------------------------------------
_import_chickens = {}, {}, ("*",) # dead chickens needed by __import__
from zope.configuration import xmlconfig

class ZCMLWrapper(object):
    """Wrapper for more convenient zcml execution."""
    auto_execute = True
    context = None
    namespaces = None
    i18n_domain = ''

    def __init__(self, context=None):
        self.context = context
        self.namespaces = {}

    def setUp(self, namespaces={}, i18n_domain=""):
        """Set active namespaces and translation domain.

           namespaces = {
               '': "http://namespaces.zope.org/zope",
               'meta': "http://namespaces.zope.org/meta"
               }
           i18ndomain = 'foo'

        Will wrap ZCML passed to string() with:

            <configure
                xmlns="http://namespaces.zope.org/zope"
                xmlns:meta="http://namespaces.zope.org/meta"
                i18ndomain="foo">
            ...
            </configure>

        """
        self.namespaces = namespaces.copy()
        self.i18n_domain = i18n_domain

    def string(self, string, name="<string>"):
        config = ''
        if self.namespaces:
            config = ' ' + ' '.join(
                ['xmlns%s="%s"' % (short and ':' + short, long)
                 for short, long in sorted(self.namespaces.items())])
        if self.i18n_domain:
            config += ' i18n_domain="%s"' % self.i18n_domain

        string = '<configure%s>\n' % config + string + '\n</configure>'

        self.context = xmlconfig.string(
            string, context=self.context, name=name,
            execute=self.auto_execute)

    def include(self, package=None, file='configure.zcml'):
        if isinstance(package, str):
            package = __import__(package, *_import_chickens)
        self.context = xmlconfig.file(
            file, package=package, context=self.context,
            execute=self.auto_execute)

    def execute(self):
        if self.context is not None:
            self.context.execute_actions()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.sampledata.name
"""

import unittest
import doctest


def doctest_NameGenerator():
    """Person name generator

    First, let's instantiate the generator, passing the random seed as
    a parameter:

        >>> from schooltool.sampledata.name import NameGenerator
        >>> gen = NameGenerator(42)

    It's got its own random generator seeded with the seed passed:

        >>> other = NameGenerator(42)
        >>> gen.random.random() == other.random.random()
        True

        >>> gen = NameGenerator(42)
        >>> other = NameGenerator(43)
        >>> gen.random.random() != other.random.random()
        True

    It's got lists of first names and last names:

        >>> import repr
        >>> print repr.repr(gen.first_names)
        ['Ada', 'Adam', 'Adrian', 'Agnieszka', 'Ainhoa', 'Al', ...]
        >>> print repr.repr(gen.last_names)
        ['Adams', 'Alexander', 'Allen', 'Alvarez', 'Andersen', 'Anderson', ...]

    Now, we can ask for a name:

        >>> gen.random.seed(42)
        >>> gen.generate()
        ('Margarita', 'Austin', 'Margarita Austin')
        >>> gen.generate()
        ('Erin', 'Duncan', 'Erin Duncan')

    The names can and will repeat:

        >>> seen = set()
        >>> count = 0
        >>> while True:
        ...     count +=1
        ...     first_name, last_name, name = gen.generate()
        ...     if name in seen:
        ...         break
        ...     else:
        ...         seen.add(name)

        >>> print count, name
        334 Joanne Snyder

    We'll have approximately this many matching pairs of names in the school:

        >>> seen = set()
        >>> for i in range(1000):
        ...     seen.add(gen.generate()[2])
        >>> 1000 - len(seen)
        3

    """

def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(optionflags=doctest.ELLIPSIS
                             |doctest.NORMALIZE_WHITESPACE),
        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve35
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.component import queryUtility, provideUtility
from zope.component.hooks import getSite, setSite
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.traversing.api import traverse
from zope.site import LocalSiteManager
from zope.site.folder import Folder
from zope.location import Location
from zope.catalog.interfaces import ICatalog

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.generations.tests import ContextStub
from schooltool.generations.evolve35 import CATALOG_KEYS


def registerLocalUtility(site, utility, name):
   manager = site.getSiteManager()
   default = traverse(site, '++etc++site/default')
   default['storage.key:'+name] = utility
   manager.registerUtility(utility, ICatalog, name)


class UtilityStub(Folder):
    def __init__(self, name):
        super(UtilityStub, self).__init__()
        self.name = name

    def __repr__(self):
        return '<%s (%s)>' % (self.__class__.__name__, self.name)


class AppStub(Folder):
    implements(ISchoolToolApplication)


def doctest_evolve35():
    """Test evolution to generation 35.

    We'll need int ids.

        >>> provideUtility(IntIds(), IIntIds)

    First, let's build ST app with local catalog utilities.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()
        >>> app.setSiteManager(LocalSiteManager(app))

        >>> for name in CATALOG_KEYS:
        ...     registerLocalUtility(app, UtilityStub(name), name)

        >>> setSite(app)
        >>> for name in CATALOG_KEYS:
        ...     print queryUtility(ICatalog, name=name, default=None)
        <UtilityStub (schooltool.basicperson)>
        <UtilityStub (schooltool.contact)>
        <UtilityStub (schooltool.person)>

    Set the site to something else and evolve.

        >>> context.root_folder['frob'] = frob = Folder()
        >>> frob.setSiteManager(LocalSiteManager(frob))
        >>> setSite(frob)

        >>> from schooltool.generations.evolve35 import evolve
        >>> evolve(context)

    Active site was kept.

        >>> getSite() is frob
        True

    And catalogs in our app were unregistered.

        >>> setSite(app)
        >>> for name in CATALOG_KEYS:
        ...     print queryUtility(ICatalog, name=name, default=None)
        None
        None
        None

    """


def setUp(test):
    setup.placefulSetUp()
    setup.setUpTraversal()


def tearDown(test):
    setup.placefulTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
schooltool.timetable unit test commons.
"""
import pytz
from datetime import datetime, date, time, timedelta

from schooltool.timetable.schedule import Meeting


class ScheduleStub(object):

    meeting_times = (
        time(0, 5),
        time(5, 0),
        time(23, 55),
        )

    def __init__(self,
                 timezone='UTC',
                 first=date(2011, 10, 29),
                 last=date(2011, 10, 31)):
        self.timezone = timezone
        self.first = first
        self.last = last

    def iterMeetings(self, start_date, until_date=None):
        if until_date is None:
            until_date = start_date

        tz = pytz.timezone(self.timezone)
        cursor = start_date
        while cursor <= until_date:
            if (cursor >= self.first and
                cursor <= self.last):
                for time_cursor in self.meeting_times:
                    starts = tz.localize(datetime.combine(cursor, time_cursor))
                    duration = timedelta(0, 900)
                    yield Meeting(starts, duration)
            cursor += cursor.resolution



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetabling Term views.
"""
import datetime
import itertools

from zope.component import adapts, getMultiAdapter
from zope.interface.exceptions import Invalid
from zope.interface import implements, directlyProvides
from zope.interface import Interface
from zope.schema import TextLine, Date
from zope.schema import ValidationError
from zope.container.interfaces import INameChooser
from zope.i18n import translate
from zope.publisher.browser import BrowserView
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.security.checker import canAccess
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form.util import getSpecification
from z3c.form.validator import NoInputData
from z3c.form.validator import WidgetsValidatorDiscriminators
from z3c.form import form, field, button
from z3c.form.validator import SimpleFieldValidator
from z3c.form.validator import WidgetValidatorDiscriminators
from z3c.form.validator import InvariantsValidator
from z3c.form.interfaces import DISPLAY_MODE
from zc.table.interfaces import ISortableColumn
from zc.table.column import GetterColumn

import schooltool.skin.flourish.breadcrumbs
from schooltool import table
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.browser.cal import month_names
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.calendar.utils import parse_date
from schooltool.calendar.utils import next_month, week_start
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.term.interfaces import ITerm, ITermContainer
from schooltool.term.term import validateTermsForOverlap
from schooltool.term.term import Term
from schooltool.skin import flourish
from schooltool.skin.dateformatter import DateFormatterMediumView
from schooltool.schoolyear.interfaces import TermOverlapError
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.common import IDateRange
from schooltool.common import DateRange
from schooltool.common import SchoolToolMessage as _
from schooltool.common.inlinept import InheritTemplate


class ITermForm(Interface):
    """Form schema for ITerm add/edit views."""

    title = TextLine(title=_("Title"))

    first = Date(title=_("Start date"))

    last = Date(title=_("End date"))


class TermFormAdapter(object):
    implements(ITermForm)
    adapts(ITerm)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __setattr__(self, name, value):
        setattr(self.context, name, value)

    def __getattr__(self, name):
        return getattr(self.context, name)


class TermView(BrowserView):
    """Browser view for terms."""

    __used_for__ = ITerm

    def calendar(self):
        """Prepare the calendar for display.

        Returns a structure composed of lists and dicts, see `TermRenderer`
        for more details.
        """
        return TermRenderer(self.context).calendar()


class FlourishTermView(flourish.page.Page, TermView, ActiveSchoolYearContentMixin):
    """flourish view of a term."""

    @property
    def title(self):
        return self.context.__parent__.title

    @property
    def subtitle(self):
        return self.context.title

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    @property
    def details(self):
        view = FlourishTermDetails(self.context, self.request, None)
        view.update()
        return view

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def done_link(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='terms')

class FlourishTermDetails(flourish.form.FormViewlet):

    fields = field.Fields(ITerm).select(
        '__name__', 'title', 'first', 'last')
    mode = DISPLAY_MODE

    def updateWidgets(self, *args, **kw):
        super(FlourishTermDetails, self).updateWidgets(*args, **kw)
        self.widgets['first'].label = _('First day')
        self.widgets['last'].label = _('Last day')


class FlourishTermActionLinks(flourish.page.RefineLinksViewlet):
    """Term action links viewlet."""


class FlourishTermsAddLinks(flourish.page.RefineLinksViewlet):
    """Term action links viewlet."""


class FlourishTermDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _(u'Delete ${term}',
                  mapping={'term': self.context.title})
        return translate(title, context=self.request)


class FlourishTermDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for confirming deletion of a schoolyear."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishTermDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishTermContainerDeleteView(flourish.containers.ContainerDeleteView, ActiveSchoolYearContentMixin):

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def nextURL(self):
        if 'CONFIRM' in self.request:
            app = ISchoolToolApplication(None)
            return self.url_with_schoolyear_id(app, view_name='terms')
        return super(flourish.containers.ContainerDeleteView, self).nextURL(self)


class TermFormBase(object):

    @property
    def showNext(self):
        return self.preview_term is None

    @property
    def showRefresh(self):
        return not self.showNext

    def setHolidays(self, term):
        term.addWeekdays(0, 1, 2, 3, 4, 5, 6)
        holidays = self.request.form.get('holiday', [])
        if not isinstance(holidays, list):
            holidays = [holidays]
        for holiday in holidays:
            try:
                term.remove(parse_date(holiday))
            except ValueError:
                pass # ignore ill-formed or out-of-range dates
        toggle = [n for n in range(7) if ('TOGGLE_%d' % n) in self.request]
        if toggle:
            term.toggleWeekdays(*toggle)


class TermAddForm(form.AddForm, TermFormBase):
    """Add form for school terms."""

    label = _("Add new term")
    template = ViewPageTemplateFile('templates/term_add.pt')

    fields = field.Fields(ITermForm)

    @property
    def preview_term(self):
        data, errors = self.extractData()
        self.updateWidgets()
        if errors:
            return None
        term = Term(data['title'], data['first'], data['last'])
        self.setHolidays(term)
        return TermRenderer(term).calendar()

    def updateActions(self):
        super(TermAddForm, self).updateActions()
        for button_id, cls in zip(['next', 'refresh', 'add'],
                                  ['button-ok', 'button-neutral', 'button-ok']):
            button = self.actions.get(button_id)
            if button is not None:
                button.addClass(cls)
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Refresh'), name='refresh',
                             condition=lambda form: form.showRefresh)
    def handleRefresh(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self._finishedAdd = False

    @button.buttonAndHandler(_('Next'), name='next',
                             condition=lambda form: form.showNext)
    def next(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self._finishedAdd = False

    @button.buttonAndHandler(_('Add term'), name='add',
                             condition=lambda form: form.showRefresh)
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        obj = self.createAndAdd(data)
        if obj is not None:
            # mark only as finished if we get the new object
            self._finishedAdd = True

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        term = Term(data['title'], data['first'], data['last'])
        form.applyChanges(self, term, data)
        self.setHolidays(term)
        return term

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def add(self, term):
        """Add `term` to the container."""
        chooser = INameChooser(self.context)
        name = chooser.chooseName("", term)
        self.context[name] = term
        return term


class FlourishTermAddView(flourish.form.AddForm, TermAddForm, ActiveSchoolYearContentMixin):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Term Details')

    @property
    def title(self):
        return self.context.title

    @button.buttonAndHandler(_('Refresh'), name='refresh',
                             condition=lambda form: form.showRefresh)
    def handleRefresh(self, action):
        super(FlourishTermAddView, self).handleRefresh.func(self, action)

    @button.buttonAndHandler(_('Next'), name='next',
                             condition=lambda form: form.showNext)
    def next(self, action):
        super(FlourishTermAddView, self).next.func(self, action)

    @button.buttonAndHandler(_('Submit'), name='add',
                             condition=lambda form: form.showRefresh)
    def handleAdd(self, action):
        super(FlourishTermAddView, self).handleAdd.func(self, action)

    def create(self, data):
        term = Term(data['title'], data['first'], data['last'])
        form.applyChanges(self, term, data)
        self.setHolidays(term)
        self._term = term
        return term

    def nextURL(self):
        return absoluteURL(self._term, self.request)

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        app = ISchoolToolApplication(None)
        url = self.url_with_schoolyear_id(app, view_name='terms')
        self.request.response.redirect(url)

    def dateString(self, date):
        return DateFormatterMediumView(date, self.request)()

    def updateWidgets(self):
        super(FlourishTermAddView, self).updateWidgets()
        description = _(u'The year starts ${year_start}',
            mapping={'year_start': self.dateString(self.context.first)})
        self.widgets['first'].field.description = description
        description = _(u'The year ends ${year_end}',
            mapping={'year_end': self.dateString(self.context.last)})
        self.widgets['last'].field.description = description


class TermEditForm(form.EditForm, TermFormBase):
    """Edit form for basic person."""
    template = ViewPageTemplateFile('templates/term_add.pt')

    fields = field.Fields(ITermForm)

    @property
    def preview_term(self):
        data, errors = self.extractData()
        self.updateWidgets()
        if errors:
            term = self.context
        else:
            term = Term(data['title'], data['first'], data['last'])
            self.setHolidays(term)
        return TermRenderer(term).calendar()

    def updateActions(self):
        super(TermEditForm, self).updateActions()
        for button_id, cls in zip(['refresh', 'apply'],
                                  ['button-neutral', 'button-ok']):
            button = self.actions.get(button_id)
            if button is not None:
                button.addClass(cls)
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Refresh'), name='refresh',
                             condition=lambda form: form.showRefresh)
    def handleRefresh(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        self._finishedAdd = False

    def applyChanges(self, data):
        changes = super(TermEditForm, self).applyChanges(data)
        self.setHolidays(self.context)
        return changes

    @button.buttonAndHandler(_('Save changes'), name='apply')
    def handleApply(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        self.status = self.successMessage

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @property
    def label(self):
        return _(u'Change information for ${term_title}',
                 mapping={'term_title': self.context.title})


class FlourishTermEditView(flourish.form.Form, TermEditForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    @property
    def title(self):
        return self.context.title

    @property
    def legend(self):
        return _(u'Change information for ${term_title}',
                 mapping={'term_title': self.context.title})

    @button.buttonAndHandler(_('Refresh'), name='refresh',
                             condition=lambda form: form.showRefresh)
    def handleRefresh(self, action):
        super(FlourishTermEditView, self).handleRefresh.func(self, action)

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishTermEditView, self).handleApply.func(self, action)
        if self.status != self.formErrorsMessage:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        super(FlourishTermEditView, self).handle_cancel_action.func(self,
            action)


class AddTermFormValidator(InvariantsValidator):

    def validateObject(self, obj):
        errors = super(AddTermFormValidator, self).validateObject(obj)
        try:
            dr = DateRange(obj.first, obj.last)
            try:
                validateTermsForOverlap(self.view.context, dr, None)
            except TermOverlapError, e:
                errors += (e, )
        except ValueError, e:
            errors += (Invalid(_("Term must begin before it ends.")), )
        except NoInputData:
            return errors
        return errors

WidgetsValidatorDiscriminators(
    AddTermFormValidator,
    view=TermAddForm,
    schema=getSpecification(ITermForm, force=True))


class EditTermFormValidator(InvariantsValidator):

    def validateObject(self, obj):
        errors = super(EditTermFormValidator, self).validateObject(obj)
        try:
            dr = DateRange(obj.first, obj.last)
            try:
                validateTermsForOverlap(self.view.context.__parent__, dr,
                                        self.view.context)
            except TermOverlapError, e:
                errors += (e, )
        except ValueError, e:
            errors += (Invalid(_("Term must begin before it ends.")), )
        except NoInputData:
            return errors
        return errors

WidgetsValidatorDiscriminators(
    EditTermFormValidator,
    view=TermEditForm,
    schema=getSpecification(ITermForm, force=True))


class DateOutOfYearBounds(ValidationError):
    """Date is not in the school year."""


class TermBoundsValidator(SimpleFieldValidator):

    def validate(self, value):
        super(TermBoundsValidator, self).validate(value)
        if ISchoolYear.providedBy(self.view.context):
            sy = self.view.context
        else:
            sy = self.view.context.__parent__

        if value not in IDateRange(sy):
            raise DateOutOfYearBounds(self.view.context, value)


class FirstTermBoundsValidator(TermBoundsValidator):
    pass

WidgetValidatorDiscriminators(
    FirstTermBoundsValidator,
    field=ITermForm['first'])


class LastTermBoundsValidator(TermBoundsValidator):
    pass

WidgetValidatorDiscriminators(
    LastTermBoundsValidator,
    field=ITermForm['last'])


class FlourishInvalidDateRangeError(ValidationError):
    __doc__ = _('Term must begin before it ends.')


class FlourishOverlapError(ValidationError):
    __doc__ = _('Date range overlaps another term.')


class FlourishOverlapValidator(TermBoundsValidator):

    def validate(self, value):
        # XXX: hack to display the overlap error next to the widget!
        super(FlourishOverlapValidator, self).validate(value)
        last_widget = self.view.widgets['last']
        last_value = self.request.get(last_widget.name)
        try:
            last_value = last_widget._toFieldValue(last_value)
        except:
            return
        try:
            dr = DateRange(value, last_value)
        except:
            raise FlourishInvalidDateRangeError()
        try:
            validateTermsForOverlap(self.container, dr, self.term)
        except TermOverlapError, e:
            raise FlourishOverlapError()


class FlourishOverlapAddValidator(FlourishOverlapValidator):
    term = None

    @property
    def container(self):
        return self.context


class FlourishOverlapEditValidator(FlourishOverlapValidator):

    @property
    def term(self):
        return self.context

    @property
    def container(self):
        return self.context.__parent__


WidgetValidatorDiscriminators(FlourishOverlapAddValidator,
                              view=FlourishTermAddView,
                              field=ITermForm['first'])


WidgetValidatorDiscriminators(FlourishOverlapEditValidator,
                              view=FlourishTermEditView,
                              field=ITermForm['first'])


class TermRenderer(object):
    """Helper for rendering ITerms."""

    first_day_of_week = 0 # Monday  TODO: get from IApplicationPreferences

    def __init__(self, term):
        self.term = term

    def calendar(self):
        """Prepare the calendar for display.

        Returns a list of month dicts (see `month`).
        """
        calendar = []
        date = self.term.first
        counter = itertools.count(1)
        while date <= self.term.last:
            start_of_next_month = next_month(date)
            end_of_this_month = start_of_next_month - datetime.date.resolution
            maxdate = min(self.term.last, end_of_this_month)
            calendar.append(self.month(date, maxdate, counter))
            date = start_of_next_month
        return calendar

    def month(self, mindate, maxdate, counter):
        """Prepare one month for display.

        Returns a dict with these keys:

            month   -- title of the month
            year    -- the year number
            weeks   -- a list of week dicts in this month (see `week`)

        """
        assert (mindate.year, mindate.month) == (maxdate.year, maxdate.month)
        weeks = []
        date = week_start(mindate, self.first_day_of_week)
        while date <= maxdate:
            weeks.append(self.week(date, mindate, maxdate, counter))
            date += datetime.timedelta(days=7)
        return {'month': month_names[mindate.month],
                'year': mindate.year,
                'weeks': weeks}

    def week(self, start_of_week, mindate, maxdate, counter):
        """Prepare one week of a Term for display.

        `start_of_week` is the date when the week starts.

        `mindate` and `maxdate` are used to indicate which month (or part of
        the month) interests us -- days in this week that fall outside
        [mindate, maxdate] result in a dict containing None values for all
        keys.

        `counter` is an iterator that returns indexes for days
        (itertools.count(1) is handy for this purpose).

        `term` is an ITerm that indicates which days are schooldays,
        and which are holidays.

        Returns a dict with these keys:

            number  -- week number
            days    -- a list of exactly seven dicts

        Each day dict has the following keys

            date    -- date as a string (YYYY-MM-DD)
            number  -- day of month
                       (None when date is outside [mindate, maxdate])
            index   -- serial number of this day (used in Javascript)
            class   -- CSS class ('holiday' or 'schoolday')
            checked -- True for holidays, False for schooldays
            onclick -- onclick event handler that calls toggle(index)

        """
        # start_of_week will be a Sunday or a Monday.  If it is a Sunday,
        # we want to take the ISO week number of the following Monday.  If
        # it is a Monday, we won't break anything by taking the week number
        # of the following Tuesday.
        week_no = (start_of_week + datetime.date.resolution).isocalendar()[1]
        date = start_of_week
        days = []
        for day in range(7):
            if mindate <= date <= maxdate:
                index = counter.next()
                checked = not self.term.isSchoolday(date)
                css_class = checked and 'holiday' or 'schoolday'
                days.append({'number': date.day, 'class': css_class,
                             'date': date.strftime('%Y-%m-%d'),
                             'index': index, 'checked': checked,
                             'onclick': 'javascript:toggle(%d)' % index})
            else:
                days.append({'number': None, 'class': None, 'index': None,
                             'onclick': None, 'checked': None, 'date': None})
            date += datetime.date.resolution
        return {'number': week_no,
                'days': days}


class FlourishTermsView(flourish.page.Page, ActiveSchoolYearContentMixin):

    @property
    def title(self):
        year = self.schoolyear
        if year is not None:
            return _('Terms for ${schoolyear}',
                     mapping={'schoolyear': year.title})

    def year(self):
        year = self.schoolyear
        if year is not None:
            return {
                'title': _(u'School Year: ${year_title}',
                         mapping={'year_title': year.title}),
                'first': year.first,
                'last': year.last,
                'empty': not bool(tuple(year.values())),
                'canModify': canAccess(year, '__delitem__'),
                'addurl': absoluteURL(year, self.request) + '/add.html',
                'alt': _(u'Add a new term to ${year_title}',
                         mapping={'year_title': year.title}),
                }

    def terms(self):
        year = self.schoolyear
        if year is not None:
            for term in reversed(tuple(year.values())):
                yield {
                    'obj': term,
                    'first': term.first,
                    'last': term.last,
                    }


class TermsTertiaryNavigationManager(
    flourish.page.TertiaryNavigationManager,
    ActiveSchoolYearContentMixin):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def items(self):
        result = []
        active = self.schoolyear
        schoolyears = active.__parent__ if active is not None else {}
        for schoolyear in schoolyears.values():
            url = '%s/terms?schoolyear_id=%s' % (
                absoluteURL(self.context, self.request),
                schoolyear.__name__)
            result.append({
                    'class': schoolyear.first == active.first and 'active' or None,
                    'viewlet': u'<a href="%s">%s</a>' % (url, schoolyear.title),
                    })
        return result


class FlourishManageYearsOverview(flourish.page.Content,
                                  ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_years_overview.pt')

    @property
    def terms(self):
        terms = ITermContainer(self.schoolyear, None)
        if terms is not None:
            return sorted(terms.values(), key=lambda t:t.first)

    def terms_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='terms')


class TermContainerBreadcrumb(flourish.breadcrumbs.Breadcrumbs, ActiveSchoolYearContentMixin):

    title = _('Terms')

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def url(self):
        if not self.checkPermission():
            return None
        url = self.url_with_schoolyear_id(self.crumb_parent, view_name='terms')
        return url

    @property
    def crumb_parent(self):
        return ISchoolToolApplication(None)


class TermAddLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        year = self.view.schoolyear
        if year is None or not flourish.canEdit(year):
            return False
        return super(TermAddLinkViewlet, self).enabled

    @property
    def url(self):
        year = self.view.schoolyear
        if year is None:
            return None
        return '%s/%s' % (absoluteURL(year, self.request), 'add.html')


def date_cell_formatter(value, item, formatter):
    view = getMultiAdapter((value, formatter.request), name='mediumDate')
    return view()


class SchoolYearTermsTable(table.ajax.Table):

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u"Title"),
            getter=lambda i, f: i.title,
            subsort=True)
        starts = GetterColumn(
            name='starts',
            title=_(u"First Day"),
            getter=lambda i, f: i.first,
            cell_formatter=date_cell_formatter,
            subsort=True)
        ends = GetterColumn(
            name='ends',
            title=_(u"Last Day"),
            getter=lambda i, f: i.last,
            cell_formatter=date_cell_formatter,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        directlyProvides(starts, ISortableColumn)
        directlyProvides(ends, ISortableColumn)
        return [title, starts, ends]

    def sortOn(self):
        return (("starts", False),)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Relationship URIs and utilities for schooltool
"""

from schooltool.course.interfaces import ICourse, ISection
from schooltool.person.interfaces import IPerson
from schooltool.group.interfaces import IGroup
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.membership import URIMembership, URIMember, URIGroup
from schooltool.relationship import URIObject, RelationshipSchema
from schooltool.relationship.interfaces import IBeforeRelationshipEvent
from schooltool.relationship.interfaces import IRelationshipAddedEvent
from schooltool.relationship.interfaces import IRelationshipRemovedEvent
from schooltool.relationship.interfaces import InvalidRelationship


#
# The Instruction relationship
#

URIInstruction = URIObject('http://schooltool.org/ns/instruction',
                           'Instruction', 'The instruction relationship.')
URISection = URIObject('http://schooltool.org/ns/instruction/section',
                       'Section', 'A role of a containing section.')
URIInstructor = URIObject('http://schooltool.org/ns/instruction/instructor',
                          'Instructor', 'A section instructor role.')

Instruction = RelationshipSchema(URIInstruction,
                                 instructor=URIInstructor,
                                 section=URISection)


def enforceInstructionConstraints(event):
    """Enforce instruction constraints (IBeforeRelationshipEvent subscriber).
    """
    if not IBeforeRelationshipEvent.providedBy(event):
        return
    if event.rel_type != URIInstruction:
        return
    if ((event.role1, event.role2) != (URIInstructor, URISection) and
        (event.role1, event.role2) != (URISection, URIInstructor)):
        raise InvalidRelationship('Instruction must have one instructor'
                                  ' and one section.')
    if not ISection.providedBy(event[URISection]):
        raise InvalidRelationship('Sections must provide ISection.')


def updateInstructorCalendars(event):
    """Add section's calendar to instructors overlaid calendars."""

    if event.rel_type != URIInstruction:
        return

    person = event[URIInstructor]
    section = event[URISection]
    calendar = ISchoolToolCalendar(section)
    if IRelationshipAddedEvent.providedBy(event):
        if calendar not in person.overlaid_calendars:
            person.overlaid_calendars.add(calendar)
    elif IRelationshipRemovedEvent.providedBy(event):
        if calendar in person.overlaid_calendars:
            person.overlaid_calendars.remove(calendar)


#
# The CourseSection relationship
#


URICourseSections = URIObject('http://schooltool.org/ns/coursesections',
                          'Course Sections',
                          'The sections that implement a course.')

URICourse = URIObject('http://schooltool.org/ns/coursesections/course',
                      'Course', 'A section that implements a course.')

# TODO it seems to me like this should be the same as URISection but because
# the URIs for roles have, so far, included the ../ns/RELATIONSHIP/.. that the
# role participates in I'm creating a seperate role URI.

URISectionOfCourse = URIObject(
                      'http://schooltool.org/ns/coursesections/section',
                      'Section', 'A course of study.')

def enforceCourseSectionConstraint(event):
    """Each CourseSections relationship requires one ICourse and one ISection
    """
    if not IBeforeRelationshipEvent.providedBy(event):
        return
    if event.rel_type != URICourseSections:
        return
    if ((event.role1, event.role2) != (URICourse, URISectionOfCourse) and
        (event.role1, event.role2) != (URISectionOfCourse, URIInstructor)):
        raise InvalidRelationship('CourseSections must have one course'
                                  ' and one section.')
    if not ISection.providedBy(event[URISectionOfCourse]):
        raise InvalidRelationship('Sections must provide ISection.')
    if not ICourse.providedBy(event[URICourse]):
        raise InvalidRelationship('Course must provide ICourse.')


CourseSections = RelationshipSchema(URICourseSections,
                                    course=URICourse,
                                    section=URISectionOfCourse)

#
# Additional handling of the section membership relationship to add sections
# to members' (students') overlay portlets automatically.  In Group membership
# relationships users must manually add the calendar to the portlet (see
# schooltool.app.membership).
#

def updateStudentCalendars(event):
    """Add section's calendar to students overlaid calendars."""

    if event.rel_type != URIMembership:
        return

    section = event[URIGroup]

    # Only continue if we're working with Sections rather than generic groups
    if not ISection.providedBy(section):
        return

    member = event[URIMember]

    calendar = ISchoolToolCalendar(section)
    if IRelationshipAddedEvent.providedBy(event):
        if IPerson.providedBy(member) and \
                calendar not in member.overlaid_calendars:
            member.overlaid_calendars.add(calendar)

        elif IGroup.providedBy(member):
            for person in member.members:
                # we don't handle nested groups any more so there
                # shouldn't be more than one layer of groups
                if IPerson.providedBy(person) and \
                        calendar not in person.overlaid_calendars:
                    person.overlaid_calendars.add(calendar)

    elif IRelationshipRemovedEvent.providedBy(event):
        if IPerson.providedBy(member):
            if calendar in member.overlaid_calendars:
                for group in member.groups:
                    if group in section.members:
                        return
                member.overlaid_calendars.remove(calendar)

        elif IGroup.providedBy(member):
            for person in member.members:
                if IPerson.providedBy(person):
                    if calendar in person.overlaid_calendars:
                        if person not in section.members:
                            person.overlaid_calendars.remove(calendar)


#
# The leadership relationship
#

URILeadership = URIObject('http://schooltool.org/ns/leadership',
                          'Leadership', 'The leadership relationship.')
URILeader = URIObject('http://schooltool.org/ns/leadership/leader',
                      'Leader', 'A role of a leader of an asset.')
URIAsset = URIObject('http://schooltool.org/ns/leadership/asset',
                     'Asset', 'An asset of a leader.')

Leadership = RelationshipSchema(URILeadership,
                                leader=URILeader,
                                asset=URIAsset)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Email browser views.

"""
import pytz
import urllib
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.i18n import translate
from zope.interface import implements, Interface, directlyProvides
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema import Text, Datetime
from zope.schema import TextLine, Int, Password, Bool
from zope.schema.interfaces import ValidationError
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.viewlet.viewlet import ViewletBase

from z3c.form import form, field, button, validator
from z3c.form.browser.checkbox import SingleCheckBoxFieldWidget
from z3c.form.interfaces import DISPLAY_MODE
from zc.table.column import GetterColumn
from zc.table.interfaces import ISortableColumn

from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.skin.containers import TableContainerView
from schooltool.skin import flourish
from schooltool.skin.flourish.form import Form
from schooltool import table
from schooltool.email.interfaces import IEmailContainer, IEmail
from schooltool.email.interfaces import IEmailUtility
from schooltool.email.mail import Email, status_messages

from schooltool.common import format_message
from schooltool.common import SchoolToolMessage as _


# Helpers

def mail_enabled():
    utility = getUtility(IEmailUtility)
    return utility.enabled()


def to_application_timezone(dt):
    app = ISchoolToolApplication(None)
    app_timezone = pytz.timezone(IApplicationPreferences(app).timezone)
    return dt.astimezone(app_timezone)


def get_application_preferences():
    app = ISchoolToolApplication(None)
    return IApplicationPreferences(app)


def set_server_status_message(form, container):
    form.widgets['server_status'].mode = 'display'
    if container.enabled:
        info = '%s:%s' % (container.hostname,
                          container.port or '25')
        form.widgets['server_status'].value = _('Enabled on ${info}',
                                                mapping={'info': info})
        form.widgets['server_status'].style = u'color: green;'
    else:
        form.widgets['server_status'].value = _('Disabled')
        form.widgets['server_status'].style = u'color: red;'


# Email Settings

class IEmailSettingsEditForm(Interface):

    server_status = TextLine(
        title=_('Server Status'),
        required=False)

    enabled = Bool(
        title=_('Enable'),
        default=False,
        required=False)

    hostname = TextLine(
        title=_('Hostname'),
        description=_('SMTP server hostname. Required if the service is enabled.'),
        required=False)

    port = Int(
        title=_('Port'),
        description=_('Port of the SMTP service. Using 25 by default.'),
        min=0,
        default=25,
        required=False)

    username = TextLine(
        title=_('Username'),
        description=_('Username used for optional SMTP authentication.'),
        required=False)

    dummy_password = TextLine(
        title=_('Password'),
        description=_('Current password to authenticate to the SMTP server.'),
        required=False)

    password = Password(
        title=_('New Password'),
        description=_('The password to authenticate to the SMTP server.'),
        required=False)

    password_confirmation = Password(
        title=_('Confirm New Password'),
        required=False)

    tls = Bool(
        title=_('TLS'),
        description=_('Use TLS connection?'),
        default=False,
        required=False)


class EmailSettingsEditFormAdapter(object):

    implements(IEmailSettingsEditForm)
    adapts(IEmailContainer)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __setattr__(self, name, value):
        if name in ('server_status', 'dummy_password',
                    'password_confirmation',):
            return
        if name in ('username',):
            if not value:
                self.context.password = None
        if name in ('password',):
            if not value:
                return
        setattr(self.context, name, value)

    def __getattr__(self, name):
        if name in ('server_status', 'dummy_password',
                    'password_confirmation'):
            return
        return getattr(self.context, name)


class EmailSettingsEditView(form.EditForm):

    form.extends(form.EditForm)
    template = ViewPageTemplateFile('templates/email_form.pt')
    label = _('Change Email Settings')
    fields = field.Fields(IEmailSettingsEditForm)
    fields['enabled'].widgetFactory = SingleCheckBoxFieldWidget

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(EmailSettingsEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def update(self):
        super(EmailSettingsEditView, self).update()
        self.updateDisplayWidgets()

    def updateDisplayWidgets(self):
        set_server_status_message(self, self.context)
        self.widgets['dummy_password'].mode = 'display'
        if self.context.password:
            mask = '*'*len(self.context.password)
            self.widgets['dummy_password'].value = mask


class FlourishEmailSettingsEditView(Form, form.EditForm):

    label = None
    legend = _('Settings')
    fields = field.Fields(IEmailSettingsEditForm)
    fields = fields.omit('server_status', 'dummy_password')

    def update(self):
        form.EditForm.update(self)

    @button.buttonAndHandler(_('Apply'))
    def handle_apply_action(self, action):
        super(FlourishEmailSettingsEditView,
              self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(FlourishEmailSettingsEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class HostnameIsRequired(ValidationError):
    __doc__ = _('Hostname is required for enabling the service')


# XXX: logic very similar to that in
#      schooltool.person.browser.person
class PasswordsDontMatch(ValidationError):
    __doc__ = _('Supplied new passwords are not identical')


class PasswordsMatchValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the validation error next to the widget!
        name = self.view.widgets['password_confirmation'].name
        verify = self.request.get(name)
        if value is not None and value not in (verify,):
            raise PasswordsDontMatch(value)


class FlourishPasswordsMatchValidator(PasswordsMatchValidator):
    pass


validator.WidgetValidatorDiscriminators(PasswordsMatchValidator,
                                        view=EmailSettingsEditView,
                                        field=IEmailSettingsEditForm['password'])


validator.WidgetValidatorDiscriminators(FlourishPasswordsMatchValidator,
                                        view=FlourishEmailSettingsEditView,
                                        field=IEmailSettingsEditForm['password'])


class HostnameValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        # XXX: hack to display the validation error next to the widget!
        name = self.view.widgets['enabled'].name
        enabled = self.request.get(name)
        if enabled and enabled[0] in ('true', 'selected') and not value:
            raise HostnameIsRequired(value)


class FlourishHostnameValidator(HostnameValidator):
    pass


validator.WidgetValidatorDiscriminators(HostnameValidator,
                                        view=EmailSettingsEditView,
                                        field=IEmailSettingsEditForm['hostname'])


validator.WidgetValidatorDiscriminators(FlourishHostnameValidator,
                                        view=FlourishEmailSettingsEditView,
                                        field=IEmailSettingsEditForm['hostname'])


# Email

class IEmailDisplayForm(Interface):

    server_status = TextLine(
        title=_('Server Status'),
        description=_('Current status of the SchoolTool email service'),
        required=False)

    from_address = TextLine(
        title=_(u'From'),
        description=_(u'The sender address'),
        required=False)

    to_addresses = TextLine(
        title=_(u'To'),
        description=_(u'Recipient addresses comma separated'),
        required=False)

    time_created = Datetime(
        title=_(u'Created on'),
        description=_(u'Date and time when the message was created'),
        required=False)

    subject = TextLine(
        title=_(u'Subject'),
        required=False)

    body = Text(
        title=_(u'Body'),
        required=False)

    status = TextLine(
        title=_(u'Email Status'),
        required=False)

    time_sent = Datetime(
        title=_(u'Last time tried'),
        description=_(u'Date and time when the message was last tried'),
        required=False)


class EmailDisplayFormAdapter(object):

    implements(IEmailDisplayForm)
    adapts(IEmail)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __getattr__(self, name):
        if name in ('server_status', 'status',):
            return
        if name == 'to_addresses':
            return ', '.join(getattr(self.context, name))
        if name in ('time_created', 'time_sent',):
            return to_application_timezone(getattr(self.context, name))
        return getattr(self.context, name)


class EmailView(form.Form):

    template = ViewPageTemplateFile('templates/email_form.pt')
    label = _('Email View')
    fields = field.Fields(IEmailDisplayForm)
    mode = 'display'

    def getDeleteURL(self, email):
        params = [('delete.%s' % (email.__name__.encode('utf-8')), ''),
                  ('CONFIRM', 'Confirm')]
        return '%s?%s' % (absoluteURL(email.__parent__, self.request),
                          urllib.urlencode(params),)

    @button.buttonAndHandler(_('Retry'))
    def handle_retry_action(self, action):
        utility = getUtility(IEmailUtility)
        email = removeSecurityProxy(self.context)
        success = utility.send(email)
        if success:
            url = self.getDeleteURL(email)
        else:
            url = absoluteURL(email.__parent__, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Delete'))
    def handle_delete_action(self, action):
        url = self.getDeleteURL(self.context)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context.__parent__, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(EmailView, self).updateActions()
        if mail_enabled():
            self.actions['retry'].addClass('button-ok')
        else:
            self.actions['retry'].mode = 'display'
        self.actions['delete'].addClass('button-cancel')
        self.actions['cancel'].addClass('button-cancel')

    def update(self):
        super(EmailView, self).update()
        self.updateDisplayWidgets()

    def updateDisplayWidgets(self):
        set_server_status_message(self, self.context.__parent__)
        if self.context.status_code is not None:
            status_text = translate(
                format_message(
                    status_messages[self.context.status_code],
                    mapping=self.context.status_parameters),
                context=self.request)
            self.widgets['status'].value = status_text
            self.widgets['status'].style = u'color: red;'


class FlourishEmailView(flourish.page.Page):

    def preferred_datetime_format(self):
        preferences = get_application_preferences()
        return '%s %s' % (preferences.dateformat, preferences.timeformat)

    def update(self):
        attrs = ['from_address', 'to_addresses', 'time_created',
                 'subject', 'body', 'status', 'time_sent']
        self.attrs = []
        datetime_format = self.preferred_datetime_format()
        info_adapter = IEmailDisplayForm(self.context)
        for attr in attrs:
            info = {'label': IEmailDisplayForm[attr].title}
            if attr != 'status':
                value = getattr(info_adapter, attr)
                if attr in ('time_created', 'time_sent'):
                    value = value.strftime(datetime_format)
            else:
                value = None
                if self.context.status_code is not None:
                    value = translate(
                        format_message(
                            status_messages[self.context.status_code],
                            mapping=self.context.status_parameters),
                        context=self.request)
            info['value'] = value
            self.attrs.append(info)

    def nextURL(self):
        return '%s/queue.html' % absoluteURL(self.context.__parent__,
                                             self.request)


class FlourishEmailRetryView(flourish.page.Page):

    def getDeleteURL(self, email):
        params = [('delete.%s' % (email.__name__.encode('utf-8')), ''),
                  ('CONFIRM', 'Confirm')]
        return '%s/%s?%s' % (absoluteURL(email.__parent__, self.request),
                             'queue.html',
                             urllib.urlencode(params),)

    def __call__(self):
        utility = getUtility(IEmailUtility)
        email = removeSecurityProxy(self.context)
        success = utility.send(email)
        if success:
            url = self.getDeleteURL(email)
        else:
            url = '%s/queue.html' % absoluteURL(self.context.__parent__,
                                                self.request)
        self.request.response.redirect(url)


# Email Container View and Table Formatter

def to_addresses_formatter(value, item, formatter):
    return ', '.join(item.to_addresses)


def subject_formatter(value, item, formatter):
    if value is None:
        return ''
    return value


def email_container_table_columns():
    from_address = GetterColumn(name='from_address',
                                title=_(u'From'),
                                getter=lambda i, f: i.from_address,
                                subsort=True)
    directlyProvides(from_address, ISortableColumn)
    to_addresses = GetterColumn(name='to_addresses',
                                title=_(u'To'),
                                getter=lambda i, f: i.to_addresses,
                                cell_formatter=to_addresses_formatter,
                                subsort=True)
    directlyProvides(to_addresses, ISortableColumn)
    subject = GetterColumn(name='subject',
                           title=_(u'Subject'),
                           getter=lambda i, f: i.subject,
                           cell_formatter=subject_formatter,
                           subsort=True)
    directlyProvides(subject, ISortableColumn)
    time_created = GetterColumn(name='time_created',
                                title=_(u'Created on'),
                                getter=lambda i, f: i.time_created,
                                cell_formatter=table.table.datetime_formatter,
                                subsort=True)
    directlyProvides(time_created, ISortableColumn)
    time_sent = GetterColumn(name='time_sent',
                                title=_(u'Last time tried'),
                                getter=lambda i, f: i.time_sent,
                                cell_formatter=table.table.datetime_formatter,
                                subsort=True)
    directlyProvides(time_sent, ISortableColumn)
    return [from_address, to_addresses, subject, time_created, time_sent]


class EmailContainerViewTableFormatter(table.table.SchoolToolTableFormatter):

    columns = lambda self: email_container_table_columns()

    def sortOn(self):
        return (('time_created', True),)


class FlourishEmailContainerViewTableFormatter(
    EmailContainerViewTableFormatter):

    def columns(self):
        from_address = GetterColumn(
            name='from_address',
            title=_(u'From'),
            getter=lambda i, f: i.from_address,
            subsort=True)
        time_created = GetterColumn(
            name='time_created',
            title=_(u'Created on'),
            getter=lambda i, f: i.time_created,
            cell_formatter=table.table.datetime_formatter,
            subsort=True)
        time_sent = GetterColumn(
            name='time_sent',
            title=_(u'Last time tried'),
            getter=lambda i, f: i.time_sent,
            cell_formatter=table.table.datetime_formatter,
            subsort=True)
        directlyProvides(from_address, ISortableColumn)
        directlyProvides(time_created, ISortableColumn)
        directlyProvides(time_sent, ISortableColumn)
        return [from_address, time_created, time_sent]


class EmailContainerView(TableContainerView):

    template = ViewPageTemplateFile('templates/email_container.pt')
    delete_template = ViewPageTemplateFile('templates/email_delete.pt')
    index_title = _('Email Queue')

    @property
    def itemsToDelete(self):
        return sorted(
            super(EmailContainerView, self)._listItemsForDeletion(),
            key=lambda obj: obj.time_created)

    def mailEnabled(self):
        return mail_enabled()

    def serverStatus(self):
        result = {}
        if self.context.hostname:
            info = '%s:%s' % (self.context.hostname,
                              self.context.port or '25')
            result['status'] = _('Enabled on ${info}',
                                 mapping={'info': info})
            result['color'] = 'green'
        else:
            result['status'] = _('Disabled')
            result['color'] = 'red'
        return result

    def __call__(self):
        if 'RETRY' in self.request:
            utility = getUtility(IEmailUtility)
            for key in self.listIdsForDeletion():
                email = removeSecurityProxy(self.context[key])
                success = utility.send(email)
                if success:
                    del self.context[key]
        return super(EmailContainerView, self).__call__()


class FlourishEmailContainerView(flourish.page.Page):
    pass


class FlourishEmailContainerDetails(flourish.form.FormViewlet):

    fields = field.Fields(IEmailSettingsEditForm)
    fields = fields.omit('server_status',
                         'enabled',
                         'password',
                         'password_confirmation')
    template = ViewPageTemplateFile("templates/f_email_container.pt")
    mode = DISPLAY_MODE

    @property
    def status(self):
        if self.context.enabled:
            return _('Enabled')
        else:
            return  _('Disabled')

    def updateWidgets(self):
        super(FlourishEmailContainerDetails, self).updateWidgets()
        if self.context.password:
            mask = '*'*len(self.context.password)
            self.widgets['dummy_password'].value = mask

    def done_link(self):
        app = ISchoolToolApplication(None)
        return '%s/settings' % absoluteURL(app, self.request)


class FlourishEmailQueueView(table.table.TableContainerView):

    def getColumnsAfter(self):
        action = table.table.ImageInputColumn(
            'delete', name='action', title=_('Delete'),
            library='schooltool.skin.flourish',
            image='remove-icon.png',
            id_getter=table.table.simple_form_key)
        return [action]

    def update(self):
        super(FlourishEmailQueueView, self).update()
        # XXX: deletion without confirmation is quite dangerous
        delete = [key for key, item in self.container.items()
                  if "delete.%s" % table.table.simple_form_key(item) in self.request]
        for key in delete:
            del self.container[key]
        if delete:
            url = absoluteURL(self.context, self.request) + '/queue.html'
            self.request.response.redirect(url)


# URL Adapter

class EmailContainerAbsoluteURLAdapter(BrowserView):

    adapts(IEmailContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/email'

    __call__ = __str__


# Action viewlets

class QueueActionMenuViewlet(ViewletBase):

    @property
    def title(self):
        count = len(self.context)
        if count:
            return _('Queue (${count} failed)', mapping={'count': count})
        else:
            return _('Queue')


# XXX: remove Send Email form?

class ISendEmailForm(Interface):

    server_status = TextLine(
        title=_('Server Status'),
        description=_('Current status of the SchoolTool email service'),
        required=False)

    from_address = TextLine(
        title=_(u'From'),
        description=_(u'The sender address'))

    to_addresses = TextLine(
        title=_(u'To'),
        description=_(u'Recipient addresses comma separated'))

    subject = TextLine(
        title=_(u'Subject'),
        description=_(u'Subject of the message'),
        required=False)

    body = Text(
        title=_(u'Body'),
        description=_(u'Body of the message'))


class SendEmailFormAdapter(object):

    implements(ISendEmailForm)
    adapts(IEmailContainer)

    def __init__(self, context):
        self.context = context


class SendEmailView(form.Form):

    template = ViewPageTemplateFile('templates/email_form.pt')
    label = _('Send Email')
    fields = field.Fields(ISendEmailForm)

    @button.buttonAndHandler(_('Send'))
    def handle_send_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        from_address = data['from_address']
        to_addresses = [address.strip()
                        for address in data['to_addresses'].split(',')]
        body = data['body']
        subject = data['subject']
        email = Email(from_address, to_addresses, body, subject)
        utility = getUtility(IEmailUtility)
        success = utility.send(email)
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(SendEmailView, self).updateActions()
        if not mail_enabled():
            self.actions['send'].mode = 'display'
        else:
            self.actions['send'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def update(self):
        super(SendEmailView, self).update()
        self.updateDisplayWidgets()

    def updateDisplayWidgets(self):
        set_server_status_message(self, self.context)


class FlourishSendEmailView(Form, SendEmailView):

    label = None
    legend = _('Email')
    fields = field.Fields(ISendEmailForm).omit('server_status')

    def update(self):
        form.Form.update(self)

    @button.buttonAndHandler(_('Send'))
    def handle_send_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        from_address = data['from_address']
        to_addresses = [address.strip()
                        for address in data['to_addresses'].split(',')]
        body = data['body']
        subject = data['subject']
        email = Email(from_address, to_addresses, body, subject)
        utility = getUtility(IEmailUtility)
        success = utility.send(email)
        if success:
            url = absoluteURL(self.context, self.request)
        else:
            url = '%s/queue.html' % absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class EmailActionsLinks(flourish.page.RefineLinksViewlet):
    """Email actions links viewlet."""


class EmailViewActionsLinks(flourish.page.RefineLinksViewlet):
    """Email view actions links viewlet."""

    body_template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/renderable_items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    # We don't want this manager rendered at all
    # if there are no renderable viewlets
    @property
    def renderable_items(self):
        result = []
        for item in self.items:
            render_result = item['viewlet']()
            if render_result and render_result.strip():
                result.append({
                        'class': item['class'],
                        'viewlet': render_result,
                        })
        return result

    def render(self):
        if self.renderable_items:
            return super(EmailViewActionsLinks, self).render()


class RetryLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return self.context.__parent__.enabled


class SendTestLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return self.context.enabled


class EmailQueueLinkViewlet(flourish.page.LinkViewlet):

    @property
    def title(self):
        count = len(self.context)
        if count:
            return _('Email Queue (${count})', mapping={'count': count})
        else:
            return _('Email Queue')


class FlourishEmailSettingsOverview(flourish.form.FormViewlet):

    fields = field.Fields(IEmailSettingsEditForm)
    fields = fields.select('server_status', 'hostname')
    template = ViewPageTemplateFile(
        'templates/f_email_settings_overview.pt')
    mode = DISPLAY_MODE

    def getContent(self):
        return IEmailContainer(self.context)

    def updateWidgets(self):
        super(FlourishEmailSettingsOverview, self).updateWidgets()
        email_container = self.getContent()
        status = email_container.enabled and _('Enabled') or _('Disabled')
        self.widgets['server_status'].value = status



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.calendar.recurrent.
"""

import time
import pytz
from datetime import datetime, date, timedelta
import unittest
import doctest

from zope.interface.verify import verifyObject


class RecurrenceRuleTestBase:
    """Base tests for the recurrence rules"""

    def test_comparison(self):
        d = self.createRule()
        d2 = d.replace()
        d3 = d.replace(count=2)
        assert d is not d2
        self.assertEqual(d, d2)
        assert not d != d2
        self.assertEqual(hash(d), hash(d2))
        assert d != None
        assert d < None or d > None
        assert d3 < d or d < d3

    def test_replace(self):
        rule = self.createRule(interval=1, until=date(2005, 1, 1))
        assert rule == rule.replace()
        rule2 = rule.replace(until=None, count=20)
        assert rule != rule2
        self.assertRaises(ValueError, rule.replace, count=20)

    def test_validate(self):
        self.assertRaises(ValueError, self.createRule, count=3,
                          until=date.today())
        self.assertRaises(ValueError, self.createRule, exceptions=(1,))
        self.assertRaises(ValueError, self.createRule, interval=0)
        self.assertRaises(ValueError, self.createRule, interval=-1)
        self.createRule(exceptions=(date.today(),))
        self.createRule(until=date.today())
        self.createRule(count=42)

    def test_iCalRepresentation(self):
        # simple case
        rule = self.createRule(interval=2)
        freq = rule.ical_freq
        self.assertEquals(rule.iCalRepresentation(None),
                          ['RRULE:FREQ=%s;INTERVAL=2' % freq])

        # count
        rule = self.createRule(interval=3, count=5)
        self.assertEquals(rule.iCalRepresentation(None),
                          ['RRULE:FREQ=%s;COUNT=5;INTERVAL=3' % freq])

        # until
        rule = self.createRule(until=date(2004, 10, 20))
        self.assertEquals(rule.iCalRepresentation(None),
                          ['RRULE:FREQ=%s;UNTIL=20041020T000000;INTERVAL=1'
                           % freq])

        # exceptions
        rule = self.createRule(exceptions=[date(2004, 10, 2*d)
                                           for d in range(3, 6)])
        self.assertEquals(rule.iCalRepresentation(None),
                          ['RRULE:FREQ=%s;INTERVAL=1' % freq,
                           'EXDATE;VALUE=DATE:20041006,20041008,20041010'])

    def test_immutability(self):
        r = self.createRule()
        for attrname in ['interval', 'count', 'until', 'exceptions']:
            self.assertRaises(AttributeError, setattr, r, attrname, 'not-ro')


class TestDailyRecurrenceRule(unittest.TestCase, RecurrenceRuleTestBase):

    def createRule(self, *args, **kwargs):
        from schooltool.calendar.recurrent import DailyRecurrenceRule
        return DailyRecurrenceRule(*args, **kwargs)

    def test(self):
        from schooltool.calendar.interfaces import IDailyRecurrenceRule
        rule = self.createRule()
        verifyObject(IDailyRecurrenceRule, rule)

    def test_apply(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(2004, 10, 13, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(2003, 10, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 21)])

        # With an end date
        rule = self.createRule(until=date(2004, 10, 20))
        result = list(rule.apply(ev))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 21)])

        # With an end date as datetime, see issue318
        rule = self.createRule(until=datetime(2004, 10, 20))
        result = list(rule.apply(ev))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 21)])

        # With a count
        rule = self.createRule(count=8)
        result = list(rule.apply(ev))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 21)])

        # With an interval
        rule = self.createRule(interval=2)
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 21, 2)])

        # With exceptions
        rule = self.createRule(exceptions=[date(2004, 10, d)
                                           for d in range(16, 21)])
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result, [date(2004, 10, d) for d in range(13, 16)])

        # With exceptions and count -- exceptions are excluded after
        # counting
        rule = self.createRule(exceptions=[date(2004, 10, d)
                                           for d in range(16, 21)],
                               count=6)
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result, [date(2004, 10, 13), date(2004, 10, 14),
                                  date(2004, 10, 15)])

        # Far in future
        tick = time.clock()
        rule = self.createRule(interval=3)
        result = list(rule.apply(ev, startdate=date(3000, 01, 01),
                                 enddate=date(3000, 01, 20)))
        self.assertEquals(int(time.clock() - tick), 0)
        self.assertEqual(result,
                         [date(3000, 1, 3), date(3000, 1, 6),
                          date(3000, 1, 9), date(3000, 1, 12),
                          date(3000, 1, 15), date(3000, 1, 18)])

    def test_scroll(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(2004, 10, 13, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # if the start date is before the event,
        # return the original event date.
        self.assertEqual(rule._scroll(ev, date(2004, 1, 1)),
                         (0, date(2004, 10, 13)))

        # if the start date coincides with the event, date
        self.assertEqual(rule._scroll(ev, date(2004, 10, 13)),
                         (0, date(2004, 10, 13)))

        # if the start date is later than the event date
        self.assertEqual(rule._scroll(ev, date(2004, 10, 23)),
                         (10, date(2004, 10, 23)))

        rule = self.createRule(interval=3)

        # if the start date is before the event,
        # return the original event date.
        self.assertEqual(rule._scroll(ev, date(2004, 1, 1)),
                         (0, date(2004, 10, 13)))

        # if the start date coincides with the event, date
        self.assertEqual(rule._scroll(ev, date(2004, 10, 13)),
                         (0, date(2004, 10, 13)))

        # if the start date is later than the event date
        self.assertEqual(rule._scroll(ev, date(2004, 10, 23)),
                         (3, date(2004, 10, 22)))

        self.assertEqual(rule._scroll(ev, date(2004, 10, 22)),
                         (3, date(2004, 10, 22)))

        self.assertEqual(rule._scroll(ev, date(2004, 10, 21)),
                         (2, date(2004, 10, 19)))


class TestYearlyRecurrenceRule(unittest.TestCase, RecurrenceRuleTestBase):

    def createRule(self, *args, **kwargs):
        from schooltool.calendar.recurrent import YearlyRecurrenceRule
        return YearlyRecurrenceRule(*args, **kwargs)

    def test(self):
        from schooltool.calendar.interfaces import IYearlyRecurrenceRule
        rule = self.createRule()
        verifyObject(IYearlyRecurrenceRule, rule)

    def test_apply(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(1970, 1, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result, [date(y, 5, 17) for y in range(1978, 2005)])

        # With an end date
        rule = self.createRule(until=date(2004, 10, 20))
        result = list(rule.apply(ev))
        self.assertEqual(result, [date(y, 5, 17) for y in range(1978, 2005)])

        # With a count
        rule = self.createRule(count=8)
        result = list(rule.apply(ev))
        self.assertEqual(result, [date(y, 5, 17) for y in range(1978, 1986)])

        # With an interval
        rule = self.createRule(interval=4)
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result,
                         [date(y, 5, 17)
                          for y in [1978, 1982, 1986, 1990, 1994, 1998, 2002]])

        # With exceptions
        rule = self.createRule(exceptions=[date(1980, 5, 17)])
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result,
                         [date(y, 5, 17)
                          for y in [1978, 1979] + range(1981, 2005)])

        # With exceptions and count -- the total nr. of events is less
        # that count.
        rule = self.createRule(exceptions=[date(1980, 5, 17)], count=4)
        result = list(rule.apply(ev, enddate=date(2004, 10, 20)))
        self.assertEqual(result,
                         [date(1978, 5, 17), date(1979, 5, 17),
                          date(1981, 5, 17)])

        # Event somewhere in the future
        rule = self.createRule(interval=2)
        result = list(rule.apply(ev, date(2000, 1, 1), date(2006, 1, 1)))
        self.assertEqual(result, [date(2000, 5, 17), date(2002, 5, 17),
                                  date(2004, 5, 17)])

        # Frebruary 29
        ev = SimpleCalendarEvent(datetime(1996, 2, 29, 12, 0),
                           timedelta(minutes=10),
                           "once in 4 years", unique_id='uid')
        rule = self.createRule()
        result = list(rule.apply(ev, date(1995, 1, 1), date(2006, 1, 1)))
        self.assertEqual(result, [date(1996, 2, 29), date(2000, 2, 29),
                                  date(2004, 2, 29)])

    def test_scroll(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')
        self.assertEqual(rule._scroll(ev, date(1900, 3, 1)),
                         (0, date(1978, 5, 17)))
        self.assertEqual(rule._scroll(ev, date(2000, 3, 1)),
                         (22, date(2000, 5, 17)))

    def test_scroll_feb29(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(1996, 2, 29, 12, 0),
                           timedelta(minutes=10),
                           "once in 4 years", unique_id='uid')
        self.assertEqual(rule._scroll(ev, date(1900, 3, 1)),
                         (0, date(1996, 2, 29)))
        self.assertEqual(rule._scroll(ev, date(1999, 3, 1)),
                         (0, date(1996, 2, 29)))
        self.assertEqual(rule._scroll(ev, date(2000, 3, 1)),
                         (4, date(2000, 2, 29)))
        self.assertEqual(rule._scroll(ev, date(2004, 3, 1)),
                         (8, date(2004, 2, 29)))


class TestWeeklyRecurrenceRule(unittest.TestCase, RecurrenceRuleTestBase):

    def createRule(self, *args, **kwargs):
        from schooltool.calendar.recurrent import WeeklyRecurrenceRule
        return WeeklyRecurrenceRule(*args, **kwargs)

    def test(self):
        from schooltool.calendar.interfaces import IWeeklyRecurrenceRule
        rule = self.createRule()
        verifyObject(IWeeklyRecurrenceRule, rule)

    def test_weeekday_validation(self):
        self.assertRaises(ValueError, self.createRule, weekdays=(1, 7))
        self.assertRaises(ValueError, self.createRule, weekdays=(1, "TH"))

    def test_replace_weekly(self):
        rule = self.createRule(weekdays=(1, 3))
        assert rule == rule.replace()
        assert rule != rule.replace(weekdays=(1,))

    def test_scroll(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        self.assertEqual(rule._scroll(ev, date(1978, 5, 25)),
                         (1, date(1978, 5, 24)))
        self.assertEqual(rule._scroll(ev, date(1978, 6, 2)),
                         (1, date(1978, 5, 24)))

        # tricky case!
        # --  Tu We
        # -- -- --
        # Mo Tu We
        # -- -- --
        # Mo Tu We
        # ...

        rule = self.createRule(weekdays=(0, 3,), interval=2)

        # We don't care that we get the closest event.  What we care
        # about is that the sequence of events is correct:
        goodresults = [(0, date(1978, 5, 17)),
                       (1, date(1978, 5, 18)),
                       (2, date(1978, 5, 29)),
                       (3, date(1978, 5, 31)),
                       (4, date(1978, 6, 1)),
                       (5, date(1978, 6, 5)),
                       (6, date(1978, 6, 7)),
                       (7, date(1978, 6, 8))]

        for delta in range(-10, 40):
            d = date(1978, 5, 17) + date.resolution * delta
            self.assert_(rule._scroll(ev, d) in goodresults, d)

        # Still, it does not always return the zeroth result:
        self.assertEqual(rule._scroll(ev, date(1978, 6, 29)),
                         (6, date(1978, 6, 14)))

    def test_apply(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(1970, 1, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(1978, 7, 17))) # Wednesday
        expected = [date(1978, 5, 17) + timedelta(w * 7) for w in range(9)]
        self.assertEqual(result, expected)

        # With an end date
        rule = self.createRule(until=date(1978, 7, 12))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an end date as datetime
        rule = self.createRule(until=datetime(1978, 7, 12))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With a count
        rule = self.createRule(count=9)
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an interval
        rule = self.createRule(interval=2, weekdays=(3,))
        result = list(rule.apply(ev, enddate=date(1978, 7, 12)))
        expected = [date(1978, 5, 17), date(1978, 5, 18),
                    date(1978, 5, 31), date(1978, 6, 1),
                    date(1978, 6, 14), date(1978, 6, 15),
                    date(1978, 6, 28), date(1978, 6, 29),
                    date(1978, 7, 12)]
        self.assertEqual(result, expected)

        # With exceptions
        rule = self.createRule(interval=2, weekdays=(3,),
                               exceptions=[date(1978, 6, 29)])
        result = list(rule.apply(ev, enddate=date(1978, 7, 12)))
        expected = [date(1978, 5, 17), date(1978, 5, 18),
                    date(1978, 5, 31), date(1978, 6, 1),
                    date(1978, 6, 14), date(1978, 6, 15),
                    date(1978, 6, 28), date(1978, 7, 12)]
        self.assertEqual(result, expected)

        # With a start date
        rule = self.createRule(interval=2, weekdays=(3,), count=9)
        result = list(rule.apply(ev, startdate=date(1978, 6, 15)))
        expected = [date(1978, 6, 15),
                    date(1978, 6, 28), date(1978, 6, 29),
                    date(1978, 7, 12)]

        self.assertEqual(result, expected)

    def test_iCalRepresentation_weekly(self):
        rule = self.createRule(weekdays=(0, 3, 6))
        dtstart = datetime(2005, 01, 01, 12, 0) # saturday
        self.assertEquals(rule.iCalRepresentation(dtstart),
                          ['RRULE:FREQ=WEEKLY;BYDAY=MO,TH,SA,SU;INTERVAL=1'])
        self.assertEquals(rule.iCalRepresentation(None),
                          ['RRULE:FREQ=WEEKLY;INTERVAL=1'])


class TestMonthlyRecurrenceRule(unittest.TestCase, RecurrenceRuleTestBase):

    def createRule(self, *args, **kwargs):
        from schooltool.calendar.recurrent import MonthlyRecurrenceRule
        return MonthlyRecurrenceRule(*args, **kwargs)

    def test(self):
        from schooltool.calendar.interfaces import IMonthlyRecurrenceRule
        rule = self.createRule()
        verifyObject(IMonthlyRecurrenceRule, rule)

    def test_monthly_validation(self):
        self.assertRaises(ValueError, self.createRule, monthly="whenever")
        self.assertRaises(ValueError, self.createRule, monthly=date.today())
        self.assertRaises(ValueError, self.createRule, monthly=None)
        self.createRule(monthly="lastweekday")
        self.createRule(monthly="monthday")
        self.createRule(monthly="weekday")

    # XXX what's with the trailing underscore?
    #     there's another test in this class, called test_replace
    def test_replace_(self):
        rule = self.createRule(monthly="lastweekday")
        assert rule == rule.replace()
        assert rule != rule.replace(monthly="monthday")

    def test_apply_monthday(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(monthly="monthday")
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(1970, 1, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(1978, 8, 17)))
        expected = [date(1978, m, 17) for m in range(5,9)]
        self.assertEqual(result, expected)

        # Over the end of the year
        result = list(rule.apply(ev, enddate=date(1979, 2, 17)))
        expected = ([date(1978, m, 17) for m in range(5, 13)] +
                    [date(1979, m, 17) for m in range(1, 3)])
        self.assertEqual(result, expected)

        # With an end date
        rule = self.createRule(monthly="monthday", until=date(1979, 2, 17))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an end datetime
        rule = self.createRule(monthly="monthday", until=datetime(1979, 2, 17))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With a count
        rule = self.createRule(count=10)
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an interval
        rule = self.createRule(monthly="monthday", interval=2)
        result = list(rule.apply(ev, enddate=date(1979, 2, 17)))
        expected = [date(1978, 5, 17), date(1978, 7, 17),date(1978, 9, 17),
                    date(1978, 11, 17), date(1979, 1, 17)]
        self.assertEqual(result, expected)

        # With exceptions
        rule = self.createRule(monthly="monthday", interval=2,
                               exceptions=[date(1978, 7, 17)])
        result = list(rule.apply(ev, enddate=date(1978, 9, 17)))
        expected = [date(1978, 5, 17), date(1978, 9, 17)]
        self.assertEqual(result, expected)

        # With a start date
        rule = self.createRule(monthly="monthday", interval=2,
                               exceptions=[date(1978, 7, 17)])
        result = list(rule.apply(ev,startdate=date(2000, 1, 1),
                                 enddate=date(2000, 6, 17)))
        expected = [date(2000, 1, 17), date(2000, 3, 17), date(2000, 5, 17)]
        self.assertEqual(result, expected)

    def test_apply_endofmonth(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(monthly="monthday")
        ev = SimpleCalendarEvent(datetime(2001, 1, 31, 0, 0),
                           timedelta(minutes=10),
                           "End of month", unique_id="uid")

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(2001, 12, 31)))
        self.assertEqual(len(result), 7)

        rule = self.createRule(monthly="monthday", count=7)
        result = list(rule.apply(ev, enddate=date(2001, 12, 31)))
        self.assertEqual(len(result), 7)
        self.assertEqual(result[-1], date(2001, 12, 31))

        rule = self.createRule(monthly="monthday", interval=2)
        result = list(rule.apply(ev, enddate=date(2002, 1, 31)))
        self.assertEqual(result, [date(2001, 1, 31),
                                  date(2001, 3, 31),
                                  date(2001, 5, 31),
                                  date(2001, 7, 31),
                                  date(2002, 1, 31),])

        result = list(rule.apply(ev, startdate=date(2001, 4, 1),
                                 enddate=date(2002, 1, 31)))
        self.assertEqual(result, [date(2001, 5, 31),
                                  date(2001, 7, 31),
                                  date(2002, 1, 31),])

    def test_apply_weekday_edge_case(self):
        # A test for Issue381 (monthly repeat causes bad date)
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(monthly="weekday")
        # 1st Tuesday and 7th day of the week
        ev = SimpleCalendarEvent(datetime(2006, 2, 7, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        result = list(rule.apply(ev, enddate=date(2006, 5, 10)))
        expected = [date(2006, 2, 7), date(2006, 3, 7),
                    date(2006, 4, 4), date(2006, 5, 2)]
        self.assertEqual(result, expected)


    def test_apply_weekday(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(monthly="weekday")
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),  # 3rd Wednesday
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(1970, 1, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(1978, 8, 17)))
        expected = [date(1978, 5, 17), date(1978, 6, 21),
                    date(1978, 7, 19), date(1978, 8, 16)]
        self.assertEqual(result, expected)

        # With a start date
        result = list(rule.apply(ev, startdate=date(1978, 8, 1),
                                 enddate=date(1979, 2, 21)))

        expected = [date(1978, 8, 16), date(1978, 9, 20),
                    date(1978, 10, 18), date(1978, 11, 15),
                    date(1978, 12, 20), date(1979, 1, 17), date(1979, 2, 21)]

        # With a start date later than recurrence on that month
        result = list(rule.apply(ev, startdate=date(1978, 8, 17),
                                 enddate=date(1979, 2, 21)))

        expected = [date(1978, 9, 20), date(1978, 10, 18), date(1978, 11, 15),
                    date(1978, 12, 20), date(1979, 1, 17), date(1979, 2, 21)]

        self.assertEqual(result, expected)
        # Over the end of the year
        result = list(rule.apply(ev, enddate=date(1979, 2, 21)))
        expected = [date(1978, 5, 17), date(1978, 6, 21),
                    date(1978, 7, 19), date(1978, 8, 16),
                    date(1978, 9, 20), date(1978, 10, 18),
                    date(1978, 11, 15), date(1978, 12, 20),
                    date(1979, 1, 17), date(1979, 2, 21)]
        self.assertEqual(result, expected)

        # With an end date
        rule = self.createRule(monthly="weekday", until=date(1979, 2, 21))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With a count
        rule = self.createRule(monthly="weekday", count=10)
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an interval
        rule = self.createRule(monthly="weekday", interval=2)
        result = list(rule.apply(ev, enddate=date(1979, 2, 21)))
        expected = [date(1978, 5, 17), date(1978, 7, 19),
                    date(1978, 9, 20), date(1978, 11, 15),
                    date(1979, 1, 17)]
        self.assertEqual(result, expected)

        # With exceptions
        rule = self.createRule(monthly="weekday", interval=2,
                               exceptions=[date(1978, 7, 19)])
        result = list(rule.apply(ev, enddate=date(1978, 9, 30)))
        expected = [date(1978, 5, 17), date(1978, 9, 20)]
        self.assertEqual(result, expected)

    def test_apply_lastweekday(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(monthly="lastweekday")
        ev = SimpleCalendarEvent(datetime(1978, 5, 17, 12, 0),  # 3rd last Wednesday
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # The event happened after the range -- empty result
        result = list(rule.apply(ev, enddate=date(1970, 1, 1)))
        self.assertEqual(result, [])

        # Simplest case
        result = list(rule.apply(ev, enddate=date(1978, 8, 17)))
        expected = [date(1978, 5, 17), date(1978, 6, 14),
                    date(1978, 7, 12), date(1978, 8, 16)]
        self.assertEqual(result, expected)

        # With a start date
        result = list(rule.apply(ev, startdate=date(1978, 8, 1),
                                 enddate=date(1979, 2, 21)))
        expected = [date(1978, 8, 16),
                    date(1978, 9, 13), date(1978, 10, 11),
                    date(1978, 11, 15), date(1978, 12, 13),
                    date(1979, 1, 17), date(1979, 2, 14)]
        self.assertEqual(result, expected)

        # With a start date later than the recurrence
        result = list(rule.apply(ev, startdate=date(1978, 8, 17),
                                 enddate=date(1979, 2, 21)))
        expected = [date(1978, 9, 13), date(1978, 10, 11),
                    date(1978, 11, 15), date(1978, 12, 13),
                    date(1979, 1, 17), date(1979, 2, 14)]
        self.assertEqual(result, expected)

        # Over the end of the year
        result = list(rule.apply(ev, enddate=date(1979, 2, 21)))
        expected = [date(1978, 5, 17), date(1978, 6, 14),
                    date(1978, 7, 12), date(1978, 8, 16),
                    date(1978, 9, 13), date(1978, 10, 11),
                    date(1978, 11, 15), date(1978, 12, 13),
                    date(1979, 1, 17), date(1979, 2, 14)]
        self.assertEqual(result, expected)

        # With an end date
        rule = self.createRule(monthly="lastweekday", until=date(1979, 2, 21))
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With a count
        rule = self.createRule(monthly="lastweekday", count=10)
        result = list(rule.apply(ev))
        self.assertEqual(result, expected)

        # With an interval
        rule = self.createRule(monthly="lastweekday", interval=2)
        result = list(rule.apply(ev, enddate=date(1979, 2, 21)))
        expected = [date(1978, 5, 17), date(1978, 7, 12),
                    date(1978, 9, 13), date(1978, 11, 15),
                    date(1979, 1, 17)]
        self.assertEqual(result, expected)

        # With exceptions
        rule = self.createRule(monthly="lastweekday", interval=2,
                               exceptions=[date(1978, 7, 12)])
        result = list(rule.apply(ev, enddate=date(1978, 9, 30)))
        expected = [date(1978, 5, 17), date(1978, 9, 13)]
        self.assertEqual(result, expected)

    def test_scroll(self):
        from schooltool.calendar.simple import SimpleCalendarEvent
        rule = self.createRule(interval=3)
        ev = SimpleCalendarEvent(datetime(2004, 10, 13, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')

        # start before event
        self.assertEqual(rule._scroll(ev, date(2004, 1, 1)),
                         (0, date(2004, 10, 13)))

        # start on event date
        self.assertEqual(rule._scroll(ev, date(2004, 10, 13)),
                         (0, date(2004, 10, 13)))

        # start after date
        self.assertEqual(rule._scroll(ev, date(2005, 3, 14)),
                         (1, date(2005, 1, 13)))

        # If we ask for an illegal dates, we get less events
        rule = self.createRule()
        ev = SimpleCalendarEvent(datetime(2004, 1, 30, 12, 0),
                           timedelta(minutes=10),
                           "reality check", unique_id='uid')
        self.assertEqual(rule._scroll(ev, date(2004, 1, 29)),
                         (0, date(2004, 1, 30)))
        self.assertEqual(rule._scroll(ev, date(2004, 2, 28)),
                         (0, date(2004, 1, 30)))
        # sequence nr 1 is taken by illegal date 2004-02-30
        self.assertEqual(rule._scroll(ev, date(2004, 3, 1)),
                         (2, date(2004, 3, 30)))

    def test_iCalRepresentation(self):
        # This method deliberately overrides the test in the base class.

        # monthday
        rule = self.createRule(monthly="monthday")
        self.assertEquals(rule.iCalRepresentation(date(2004, 10, 26)),
                          ['RRULE:FREQ=MONTHLY;BYMONTHDAY=26;INTERVAL=1'])

        # weekday
        rule = self.createRule(monthly="weekday")
        self.assertEquals(rule.iCalRepresentation(date(2004, 10, 26)),
                          ['RRULE:FREQ=MONTHLY;BYDAY=4TU;INTERVAL=1'])

        # lastweekday
        rule = self.createRule(monthly="lastweekday")
        self.assertEquals(rule.iCalRepresentation(date(2004, 10, 26)),
                          ['RRULE:FREQ=MONTHLY;BYDAY=-1TU;INTERVAL=1'])

        # some standard stuff
        rule = self.createRule(interval=3, count=7,
                               exceptions=[date(2004, 10, 2*d)
                                           for d in range(3, 6)])
        self.assertEquals(rule.iCalRepresentation(date(2004, 10, 26)),
                      ['RRULE:FREQ=MONTHLY;COUNT=7;BYMONTHDAY=26;INTERVAL=3',
                       'EXDATE;VALUE=DATE:20041006,20041008,20041010'])


class TestWeekSpan(unittest.TestCase):

    def test_weekspan(self):
        from schooltool.calendar.recurrent import weekspan

        # The days are in the same week
        self.assertEqual(weekspan(date(2004, 10, 11), date(2004, 10, 17)), 0)
        #                              Monday, w42         Sunday, w42

        # The days are in the adjacent weeks
        self.assertEqual(weekspan(date(2004, 10, 17), date(2004, 10, 18)), 1)
        #                              Sunday, w42         Monday, w43

        # The days span the end of year
        self.assertEqual(weekspan(date(2004, 12, 30), date(2005, 01, 07)), 1)
        #                              Thursday, w53       Friday, w1

        # The days span the end of year, two weeks
        self.assertEqual(weekspan(date(2004, 12, 30), date(2005, 01, 14)), 2)
        #                              Thursday, w53       Friday, w2


class TestMonthIndex(unittest.TestCase):

    def test_monthindex(self):
        from schooltool.calendar.recurrent import monthindex
        # First Friday of October 2004
        self.assertEqual(monthindex(2004, 10, 1, 4), date(2004, 10, 1))
        self.assertEqual(monthindex(2004, 10, 1, 3), date(2004, 10, 7))
        self.assertEqual(monthindex(2004, 10, 1, 3), date(2004, 10, 7))

        # Users must check whether the month is correct themselves.
        self.assertEqual(monthindex(2004, 10, 5, 3), date(2004, 11, 4))

        self.assertEqual(monthindex(2004, 10, 4, 3), date(2004, 10, 28))
        self.assertEqual(monthindex(2004, 10, -1, 3), date(2004, 10, 28))

        self.assertEqual(monthindex(2004, 11, -1, 1), date(2004, 11, 30))
        self.assertEqual(monthindex(2004, 11, -1, 2), date(2004, 11, 24))

        self.assertEqual(monthindex(2004, 12, -1, 3), date(2004, 12, 30))
        self.assertEqual(monthindex(2004, 12, -1, 4), date(2004, 12, 31))
        self.assertEqual(monthindex(2004, 12, -1, 3), date(2004, 12, 30))
        self.assertEqual(monthindex(2004, 12, -2, 3), date(2004, 12, 23))


def doctest_InfinitePastEventsBug():
    """Regression test for http://issues.schooltool.org/issue461

        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from schooltool.calendar.recurrent import WeeklyRecurrenceRule
        >>> ev = SimpleCalendarEvent(datetime(2006, 1, 18, tzinfo=pytz.utc),
        ...                          timedelta(hours=1), 'Sample event',
        ...                          recurrence=WeeklyRecurrenceRule())
        >>> for e in ev.expand(datetime(2006, 1, 1, tzinfo=pytz.utc),
        ...                    datetime(2006, 2, 1, tzinfo=pytz.utc)):
        ...     print e.dtstart
        2006-01-18 00:00:00+00:00
        2006-01-25 00:00:00+00:00

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite())
    suite.addTest(unittest.makeSuite(TestDailyRecurrenceRule))
    suite.addTest(unittest.makeSuite(TestYearlyRecurrenceRule))
    suite.addTest(unittest.makeSuite(TestWeeklyRecurrenceRule))
    suite.addTest(unittest.makeSuite(TestMonthlyRecurrenceRule))
    suite.addTest(unittest.makeSuite(TestWeekSpan))
    suite.addTest(unittest.makeSuite(TestMonthIndex))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
